"""Analysis of market instruments.

Analysis is provides for by the methods of the analysis classes `Analysis`
and `Compare`. The `Analysis` class provides for analysis of a single
instrument whilst the `Compare` class provides for analysis of multiple
symbols.

The symbols to be analysed are passed to the analysis class together with,
optionally, a `market_prices` class to use to get price data. By default
this class will be `market_prices.PricesYahoo`.

** Parameters defining analyis period **
The period to be analysed is passed to the methods of the analysis class
as period parameters that will be passed on to the 'get' method of the
`market_prices` class, i.e. by default to `PricesYahoo.get`. Call help on
this underlying method for doc on the available prices parameters, for
example with `help(PricesYahoo.get)`, or directly from an analysis instance
with `help(<analysis_instance>.prices.get)`. The analysis method will also
pass on all other parameters save for those that might be defined by the
method itself (as noted in the method doc).

If period parameters are not defined then the analysis will usually be
undertaken over the longest period for which data is available.

** 'style' argument **
Some analysis methods offer a 'style' which if passed as True (default)
will format the ouput and return as Styler. To return as the underlying
data pass 'style' as False.

Functions
---------
pct_chg_top_to_bot(): -> pd.Series | float:
    Percentage change between first and last value of pandas object.

abs_chg_top_to_bot(): -> pd.Series | float:
    Absolute change between first and last value of pandas object.

add_summary(): -> pd.DataFrame
    Add summary row and/or column (same agg func applied to all rows/cols).

add_summary_row(): -> pd.DataFrame
    Add summary row commposed of different agg funcs.

max_advance() -> pd.DataFrame:
    Maximum percentage advance.

max_decline() -> pd.DataFrame:
    Maximum percentage decline.

formatter_datetime():
    Format value to format suitable for timestamp accuracy.

formatter_percent(x: float):
    Format value as percentage.

formatter_float(x: float):
    Format value as float.

style_df(): -> Styler
    Styler with columns styled according to `FORMATTERS`

style_df_grid(): -> Styler
    Styler in grid style.

Classes
-------
Base(metaclass=ABCMeta):
    Base functionality for analysis of one or more instruments.

Analysis(Base):
    Analyse single instrument.

Compare(Base):
    Analyse multiple insturments.

TestAnalysis(_TestBase):
    Test for Analysis class.

TestCompare(_TestBase):
    Test for Compare class.
"""

from __future__ import annotations

from abc import ABCMeta
from collections.abc import Callable, Sequence
from datetime import datetime
from typing import TYPE_CHECKING, Any, Literal

import market_prices as mp
import matplotlib as mpl
import pandas as pd
from exchange_calendars import ExchangeCalendar
from pandas.io.formats.style import Styler

from market_analy import guis
from market_analy.formatters import FORMATTERS, formatter_percent
from market_analy.utils.mkt_prices_utils import (
    period_string,
    range_string,
    request_daily_prices,
)
from market_analy.utils.pandas_utils import (
    rebase_to_row,
    interval_index_new_tz,
    index_dates_to_str,
)

if TYPE_CHECKING:
    from .trends.movements import MovementsSupportChartAnaly
    from .trends import TrendsProto

from .trends.guis import TrendsGui, TrendsGuiBase
from .trends.analy import Trends

Date = str | datetime | None
Calendar = str | ExchangeCalendar
Symbols = str | list[str]
ChartEngine = Literal["bqplot", "matplotlib"]


def pct_chg_top_to_bot(pd_obj: pd.Series | pd.DataFrame) -> pd.Series | float:
    """Return pct change between first and last value of `pd_obj`.

    Returns
    -------
    pd.Series | float
        If `pd_obj` a `pd.Series` then returns float, if `pd.DataFrame`
        then returns `pd.Series` of float values.
    """
    return pd_obj.iloc[-1] / pd_obj.iloc[0] - 1


def abs_chg_top_to_bot(pd_obj: pd.Series | pd.DataFrame) -> pd.Series | float:
    """Return abs change between first and last value of `pd_obj`.

    Returns
    -------
    pd.Series | float
        If `pd_obj` a pd.Series then returns float, if pd.DataFrame returns
        Series of float values.
    """
    return pd_obj.iloc[-1] - pd_obj.iloc[0]


def add_summary(
    df: pd.DataFrame,
    aggmethod: str = "mean",
    axis: Literal["row", "column", "both"] = "row",
    label: str = "SUMMARY",
) -> pd.DataFrame:
    """Add summary row and/or column.

    Parameters
    ----------
    aggmethod
        `pd.DataFrame` method to be used to calcuale summary. Will be
        applied to all rows/columns. NB use `add_summary_row` to create
        a summary row composed of different aggregate methods for different
        columns.

    axis
        Direction summary being applied.

    label
        Label for summary row/colum.
    """
    if axis not in (opts := ["row", "column", "both"]):
        msg = f"`axis` invalid, received as {axis} although must be in {opts}."
        raise ValueError(msg)
    if axis != "column":
        srs = getattr(df, aggmethod)(axis=0)
        srs.name = label
        summary_row = pd.DataFrame(srs).transpose()
    if axis != "row":
        summary_column = getattr(df, aggmethod)(axis=1)
        summary_column.name = label

    if axis == "row":
        return pd.concat((df, summary_row), axis=0)
    elif axis == "column":
        return pd.concat([df, summary_column], axis=1)
    rtrn = pd.concat((df, summary_row), axis=0)
    return pd.concat([rtrn, summary_column], axis=1)


SummaryRowFuncDef = tuple[str, str | list[str]]
SummaryRowDef = SummaryRowFuncDef | list[SummaryRowFuncDef]


def add_summary_row(
    df: pd.DataFrame,
    summary: SummaryRowDef = ("mean", "pct_chg"),
    label: str = "SUMMARY",
) -> pd.DataFrame:
    """Add summary row commposed of multiple aggregating functions.

    NB use `add_summary` to add a summary row composed of a single
    aggregating function.

    Parameters
    ----------
    summary
        2-tuple or list of 2-tuples where tuple takes:
            [0] string of the name of an aggregating method that is to
            contribute to populating the summary row.

            [1] string of column name, or list of column names, to which
            aggregating method to be applied.

    label:
        Label for summary row.
    """
    kwargs = {}
    summary = [summary] if isinstance(summary, tuple) else summary
    for s in summary:
        cols = s[1] if isinstance(s[1], list) else [s[1]]
        for col in cols:
            kwargs.update({col: pd.NamedAgg(column=col, aggfunc=s[0])})
    summary_row = df.groupby(by=lambda _: label).agg(**kwargs)
    return pd.concat((df, summary_row))


DURATION_COLUMNS = ["days", "hours", "minutes"]


def _add_duration_columns(d: dict, start: pd.Timestamp, end: pd.Timestamp) -> dict:
    diff = end - start
    for attr in DURATION_COLUMNS:
        v = getattr(diff.components, attr)
        if v > 0:
            d[attr] = [v]
    return d


def max_advance(prices: pd.DataFrame, label: Any = None) -> pd.DataFrame:
    """Return maximum percentage advance.

    Parameters
    ----------
    prices
        Price data covering period within which to return maximum advance.

    label
        Value to assign to index label. Default 0.
    """
    df = prices.loc[:, ["high", "low"]]

    df["max_adv"] = df["high"] / df["low"].cummin() - 1
    max_adv = df["max_adv"].max()

    end = df[df["max_adv"] == max_adv]
    end = end.index[-1]  # if >1 max adv of same value, the most recent

    df = df.loc[:end]
    start = df[df["low"] == df["low"].min()]
    start = start.index[-1]  # if >1 lows of same value, the most recent

    d = {
        "start": [start],
        "low": [df.loc[start, "low"]],
        "end": [end],
        "high": [df.loc[end, "high"]],
        "pct_chg": [max_adv],
    }
    d = _add_duration_columns(d, start, end)
    return pd.DataFrame(d, index=[label])


def max_decline(prices: pd.DataFrame, label: Any = None) -> pd.DataFrame:
    """Return maximum percentage decline.

    Parameters
    ----------
    prices
        Price data covering period within which to return maximum decline.

    label
        Value to assign to index label. Default 0.
    """
    px = prices.loc[:, ["high", "low"]]

    px["max_dec"] = px["low"] / px["high"].cummax() - 1
    max_dec = px["max_dec"].min()

    end = px[px["max_dec"] == max_dec]
    end = end.index[-1]  # if >1 max dec of same value, the most recent

    px = px.loc[:end]
    start = px[px["high"] == px["high"].max()]
    start = start.index[-1]  # if >1 highs of same value, the most recent

    d = {
        "start": [start],
        "high": [px.loc[start, "high"]],
        "end": [end],
        "low": [px.loc[end, "low"]],
        "pct_chg": [max_dec],
    }
    d = _add_duration_columns(d, start, end)
    return pd.DataFrame(d, index=[label])


def _color_chg(value: float | int) -> str:
    """Return colour to represent sign of a value.

    Returns
    -------
        Css property:
            "color: crimson" for negative values.
            "color: yellowgreen" for negative values.
    """
    color = "Crimson" if value < 0 else "YellowGreen"
    return f"color: {color}"


def style_df(
    df: pd.DataFrame,
    na_rep: Any = "",
    chg_cols: Sequence[str] = ("chg", "pct_chg"),
    caption: str | None = None,
) -> Styler:
    """Style `pd.DataFrame` with columns included to `FORMATTERS`."""
    formatter = {k: v for k, v in FORMATTERS.items() if k in df.columns}
    styler = df.style.format(formatter, na_rep=na_rep)

    chg_subset = [c for c in df.columns if c in chg_cols]
    if chg_subset:
        styler.map(_color_chg, subset=chg_subset)

    if caption:
        styler.set_caption(caption)

    return styler


def style_df_grid(
    df: pd.DataFrame,
    format: str | None = "{:.2%}",
    na_rep: Any = "",
    gradient_kwargs: dict | None = {"cmap": "RdBu"},
    caption: str | None = None,
) -> Styler:
    """Style `df` in a grid style."""
    styler = df.style
    if format is not None:
        styler.format(format, na_rep=na_rep)
    if gradient_kwargs is not None:
        styler.background_gradient(axis=None, **gradient_kwargs)
    if caption is not None:
        styler.set_caption(caption)
    return styler


# Defaults
CHG_SUMMARY_PERIODS = [
    {"minutes": 5},
    {"minutes": 15},
    {"hours": 1},
    {"hours": 4},
    {"days": 1},
    {"days": 2},
    {"days": 5},
    {"weeks": 2},
    {"months": 1},
    {"months": 3},
    {"months": 6},
    {"years": 1},
]


# Classes
class Base(metaclass=ABCMeta):
    """Base functionality for analysis of one or more instruments.

    Properties
    ----------
    prices : PricesBase, default: PricesYahoo
        Prices accessor.

    today -> Styler:
        Price row for most recent day.

    now -> pd.DataFrame:
        Most recently registered price.

    Methods
    -------
    daily_prices() -> pd.DataFrame
        Daily open, high, low, close, volume.

    daily_close_prices() -> pd.DataFrame
        Daily close prices.

    pct_chg() -> pd.DataFrame | Styler:
        Percentage change over specific period.

    summary_chg() -> pd.DataFrame:
        Percentage change over multiple periods.

    price_chg() -> pd.Series | float:
        Absolute price change over specified period.

    chg() -> pd.DataFrame | Styler:
        Absolute and percentage price change over specified period.

    chg_every_interval() -> pd.DataFrame | Styler:
        Percent change each interval over specified period.

    price_on() -> pd.DataFrame | Styler:
        Price row for a day. Most recent by default.

    price_at() -> pd.DataFrame:
        Price at a specific datetime.

    price() -> pd.DataFrame:
        Price on a day or at a timestamp.

    price_range() -> pd.DataFrame:
        OHLCV over a period.
    """

    PctChgBarCls = guis.GuiPctChg

    def __init__(self, prices: mp.PricesBase):
        """`analysis.Base` constructor.

        Parameters
        ----------
        prices
            Instance of a subclass of `market_prices.PricesBase` from
            which to request price data for the symbols to be analysed.
            For example:
                prices = market_prices.PricesYahoo("MSFT")
                prices = market_prices.PricesYahoo("MSFT, AZN.L")
        """
        self._prices = prices

    @property
    def prices(self) -> mp.PricesBase:
        """Associated prices object."""
        return self._prices

    def _daily_prices(self, close_only: bool, **kwargs) -> pd.DataFrame:
        kwargs["close_only"] = close_only
        kwargs["interval"] = "1d"
        kwargs.setdefault("lose_single_symbol", True)
        return self.prices.get(**kwargs)

    def daily_prices(self, **kwargs) -> pd.DataFrame:
        """Daily open, high, low, close, volume.

        If multiple symbols, columns indexed with MultiIndex with symbol
        in level 0.

        Parameters
        ----------
        **kwargs
            Parameters to define period / price data to be analysed. See
            module doc with `help(analysis.__doc__)`. Cannot include
            'close_only' or 'interval'.
        """
        return self._daily_prices(close_only=False, **kwargs)

    def daily_close_prices(self, **kwargs) -> pd.DataFrame:
        """Daily close prices.

        Columns indexed with symbol only. Not multiindexed.

        Parameters
        ----------
        **kwargs
            Parameters to define period / price data to be analysed. See
            module doc with `help(analysis.__doc__)`. Cannot include
            'close_only' or 'interval'.
        """
        return self._daily_prices(close_only=True, **kwargs)

    def price_chg(self, style: bool = True, **kwargs) -> pd.Series | float:
        """Absolute price change over specified period.

        Parameters
        ----------
        **kwargs
            Parameters to define period / price data to be analysed. See
            module doc with `help(analysis.__doc__)`. Cannot include
            'interval', 'add_a_row', 'close_only' or 'lose_single_symbol'
        """
        interval = "1D" if request_daily_prices(**kwargs) else None
        df = self.prices.get(
            interval=interval,
            add_a_row=True,
            fill="both",
            close_only=True,
            lose_single_symbol=False,
            **kwargs,
        )
        abs_chg = abs_chg_top_to_bot(df)
        res = pd.DataFrame({"chg": abs_chg})
        if style:
            caption = "Price Change " + range_string(df.index, close=True)
            return style_df(res, caption=caption)
        else:
            return res

    def pct_chg(self, style: bool = True, **kwargs) -> pd.DataFrame | Styler:
        """Percentage change over specified period.

        Parameters
        ----------
        **kwargs
            Parameters to define period / price data to be analysed. See
            module doc with `help(analysis.__doc__)`. Cannot include
            'interval', 'add_a_row', 'close_only' or 'lose_single_symbol'.
        """
        interval = "1D" if request_daily_prices(**kwargs) else None
        df = self.prices.get(
            interval=interval,
            add_a_row=True,
            fill="both",
            close_only=True,
            lose_single_symbol=False,
            **kwargs,
        )
        pct_chg = pct_chg_top_to_bot(df)
        res = pd.DataFrame({"pct_chg": pct_chg})
        if style:
            if len(res.index) > 1:
                res = add_summary_row(res, ("mean", "pct_chg"), label="Av.")
            range_str = range_string(df.index, close=True)
            caption = "Percentage Change " + range_str
            return style_df(res, caption=caption)
        else:
            return res

    def summary_chg(
        self, periods: list[dict] | None = None, style: bool = True
    ) -> pd.DataFrame:
        """Percentage change over multiple periods.

        Parameters
        ----------
        periods
            List of dictionaries with each dictionary defining a period
            over which require pct change. Period defined with period
            parameters. See module doc with `help(analysis.__doc__)`.
            Cannot include 'interval', 'add_a_row', 'close_only' or
            'lose_single_symbol'.
        """
        periods = CHG_SUMMARY_PERIODS if periods is None else periods
        dfs = [self.pct_chg(style=False, **period) for period in periods]
        df = pd.concat(dfs, axis=1)
        df.columns = [period_string(**period) for period in periods]
        if style:
            if len(df.index) > 1:
                df = add_summary(df, "mean", axis="row", label="Av.")
            styler = style_df(
                df, chg_cols=list(df.columns), caption="Percentage Change"
            )
            styler.format(formatter_percent)
            return styler
        else:
            return df

    def chg(self, style: bool = True, **kwargs) -> pd.DataFrame | Styler:
        """Absolute and percentage price change over specified period.

        Parameters
        ----------
        **kwargs
            Parameters to define period / price data to be analysed. See
            module doc with `help(analysis.__doc__)`. Cannot include
            'interval', 'add_a_row', 'close_only' or 'lose_single_symbol'.
        """
        interval = "1D" if request_daily_prices(**kwargs) else None
        df = self.prices.get(
            interval=interval,
            add_a_row=True,
            fill="both",
            close_only=True,
            lose_single_symbol=False,
            **kwargs,
        )
        abs_chg = abs_chg_top_to_bot(df)
        pct_chg = pct_chg_top_to_bot(df)
        res = pd.DataFrame({"pct_chg": pct_chg, "chg": abs_chg, "close": df.iloc[-1]})
        if style:
            if len(res.index) > 1:
                res = add_summary_row(res, ("mean", "pct_chg"), label="Av.")
            caption = "Change " + range_string(df.index, close=True)
            return style_df(res, caption=caption)
        return res

    def chg_every_interval(
        self,
        interval: str = "1d",
        style: bool = True,
        chart: bool = False,
        direction: Literal["horizontal", "vertical"] = "vertical",
        _display: bool = True,
        **kwargs,
    ) -> pd.DataFrame | Styler:
        """Percent change every `interval` over specified period.

        Parameters
        ----------
        interval
            Interval over which to evaluate each percentage changes. Passed
            as interval parameter of `self.prices.get`, examples:
                '5T', '15T', '1h', '1d', '5d', '1M'

        style
            True to format ouput and return as Styler.
            Ignored if chart True.

        chart
            True to display as Bar chart.

        direction
            (only relevant is `chart` True)
            Direction of bars.

        **kwargs
            Parameters to define period / price data to be analysed. See
            module doc with `help(analysis.__doc__)`. Cannot include
            'add_a_row' or 'close_only'.

        _display
            (for testing only)
            Do not display chart when chart is True.
        """
        df = self.prices.get(
            interval,
            close_only=True,
            add_a_row=True,
            fill="both",
            **kwargs,
        )
        shifted = df.shift(1)
        chgs = (df - shifted) / shifted
        chgs = chgs[1:]
        if chart or style:
            caption = (
                "Change over prior "
                + interval
                + ".   Period: "
                + period_string(**kwargs)
            )

        if chart:
            return self.PctChgBarCls(
                data=chgs, title=caption, direction=direction, display=_display
            )

        elif style:
            if isinstance(chgs.index, pd.DatetimeIndex):
                chgs.index = index_dates_to_str(chgs.index)
            else:
                chgs.index = interval_index_new_tz(chgs.index, None)
            symbol_cols = chgs.columns
            styler = style_df(chgs.reset_index(), chg_cols=symbol_cols, caption=caption)
            styler.format({c: formatter_percent for c in symbol_cols})
            styler.hide(axis="index")
            return styler
        return chgs

    def price_on(
        self, session: pd.Timestamp | str | None = None, style: bool = True
    ) -> pd.DataFrame | Styler:
        """Price row for a session.

        Parameters
        ----------
        session: pd.Timestamp, default: Most recent for `.prices` lead symbol.
            Session for which to return price row, for example
            "2023-01-05".

        style: bool, default: True
            True: Format output (returns `Styler`).
            False: Return underlying `pd.DataFrame`.
        """
        if session is None:
            cal = self.prices.calendar_default
            symbol = self.prices.lead_symbol_default
            minute = pd.Timestamp.now() - self.prices.delays[symbol]
            session = cal.minute_to_session(minute, "previous")
        df = self.prices.session_prices(session)
        date = df.index[0]
        df = df.pt.stacked.droplevel(0)  # one row for each symbol
        chg_df = self.chg(end=date, days=1, style=False)
        chg_df.pop("close")  # so as not to replicate column
        df = pd.concat([df, chg_df], axis=1)
        df = df[["pct_chg", "chg", "close", "open", "high", "low", "volume"]]
        if not style:
            return df
        if len(df.index) > 1:
            df = add_summary_row(df, ("mean", "pct_chg"), label="Av.")
        caption = f"Price {date.strftime('%Y-%m-%d')}"
        return style_df(df, caption=caption)

    def price_at(self, ts: pd.Timestamp | str | None = None) -> pd.DataFrame:
        """Price at a specific datetime.

        Parameters
        ----------
        ts
            Datetime at which to return most recently registered price.

            Default: most recently registered price at lowest available
            interval.
        """
        return self.prices.price_at(ts)

    def price(
        self, ts: pd.Timestamp | str | None = None, style: bool = True
    ) -> pd.DataFrame:
        """Price on a day or at a timestamp.

        Parameters
        ----------
        ts: pd.Timestamp | str | None, default: most recent day
            If require price for a day then `ts` should be passed with no
            time component and with tz as None.

            Otherwise, timestamp for which require most recently registered
            price.

        style
            True to format output and return a `Styler`.
            Only relevant if label represents a day.
        """
        if ts is None or mp.helpers.is_date(pd.Timestamp(ts)):
            return self.price_on(ts, style)
        else:
            return self.price_at(ts)

    def price_range(self, **kwargs) -> pd.DataFrame:
        """OHLCV over period.

        Exposes `self.prices.price_range`.

        See `help(self.prices.price_range)` for parameters.
        """
        return self.prices.price_range(**kwargs)

    @property
    def today(self) -> Styler:
        """Price row for most recent day."""
        return self.price_on(style=True)

    @property
    def now(self) -> pd.DataFrame:
        """Most recently registered price."""
        return self.price_at()


class Analysis(Base):
    """Single instrument analysis.

    Properties
    ----------

    symbol -> str:
        Instrument symbol.

    Methods
    -------
    plot():
        Chart prices over a specified period.

    max_adv() -> pd.Dataframe | Styler:
        Maximum percentage advance over specified period.

    max_dec() -> pd.Dataframe | Styler:
        Maximum percentage decline over specified period.

    corr() -> pd.DataFrame | Styler:
        Correlation with other Analysis object over specified period.
    """

    def __init__(self, prices: mp.PricesBase):
        """Construct `Analysis` instance.

        Parameters
        ----------
        prices:
            Instance of a subclass of `market_prices.PricesBase` from
            which to request price data for the single symbol to be
            analysed. For example:
                prices = market_prices.PricesYahoo("MSFT")

        Raises
        ------
        ValueError
            If `prices` gets prices for multiple symbols.
        """
        if len(prices.symbols) > 1:
            msg = (
                "The Analysis class requires a `prices` instance that gets"
                "price data for a single symbol, although the past instance"
                f" gets prices for {len(prices.symbols)}: {prices.symbols}."
            )
            raise ValueError(msg)
        super().__init__(prices)

    @property
    def symbol(self) -> str:
        """Instrument symbol."""
        return self.prices.symbols[0]

    def plot(
        self,
        interval: mp.intervals.RowInterval | None = None,
        engine: ChartEngine = "bqplot",
        chart_type: Literal["line", "candle"] = "candle",
        max_ticks: int | None = None,
        log_scale: bool = True,
        **kwargs,
    ) -> guis.GuiOHLC | guis.GuiLine | mpl.artist.Artist:
        """Chart prices over specified period.

        Parameters
        ----------
        interval
            Price data interval to use, as 'interval` parameter described
            by `help(self.prices.get)`.

            Default: None if `engine` "bqplot" otherwise '1d'.

        engine : Literal["bqplot", "matplotlib"]
            Chart backend.

        chart_type
            Chart type. 'candle' option only implemented for bqplot engine.

        log_scale
            True for price axis to be to log scale. Only implemented if
            `engine` is "bqplot".

        max_ticks
            Maximum number of x-axis ticks that will shown by default (more
            can be shown via slider). None for no limit. Only implemented
            if `engine` is "bqplot".

        **kwargs:
            Parameters to define period / price data to be analysed. See
            method doc with `help(analysis.__doc__)`. Cannot include
            'close_only' if engine is "matplotlib".
        """
        if engine == "bqplot":
            kwargs["interval"] = interval
            Cls = guis.GuiOHLC if chart_type == "candle" else guis.GuiLine
            return Cls(self, log_scale=log_scale, max_ticks=max_ticks, **kwargs)
        else:
            interval = "1d" if interval is None else interval
            subset = self.prices.get(close_only=True, **kwargs)
            return subset.plot()

    def max_adv(self, style: bool = True, **kwargs) -> pd.DataFrame | Styler:
        """Maximum percentage advance over specified period.

        Parameters
        ----------
        **kwargs
            Parameters to define period / price data to be analysed. See
            method doc with `help(analysis.__doc__)`. Cannot include
            'lose_single_symbol'.

            By default return will be evaluted from data of the highest
            interval that can most accurately reflect the period start and
            end (either as passed or evaluated). This will be a composite
            table if the 'end' could be represented more accurately by a
            lower interval for which data is not otherwise available to
            cover the full requested period. Alternatively, 'interval' can
            be passed to evaluate the movement based on data for a specific
            interval.
        """
        if "interval" not in kwargs:
            kwargs["composite"] = True
        prices = self.prices.get(lose_single_symbol=True, fill="both", **kwargs)
        if isinstance(prices.index, pd.IntervalIndex):
            prices = prices.pt.indexed_left
        df = max_advance(prices, label=self.symbol)
        if not style:
            return df
        caption = f"Maximum Advance {period_string(**kwargs)}"
        return style_df(df, caption=caption)

    def max_dec(self, style: bool = True, **kwargs) -> pd.DataFrame | Styler:
        """Maximum percentage decline over specified period.

        Parameters
        ----------
        **kwargs
            Parameters to define period / price data to be analysed. See
            method doc with `help(analysis.__doc__)`. Cannot include
            'lose_single_symbol'.

            By default return will be evaluted from data of the highest
            interval that can most accurately reflect the period start and
            end (either as passed or evaluated). This will be a composite
            table if the 'end' could be represented more accurately by a
            lower interval for which data is not otherwise available to
            cover the full requested period. Alternatively, 'interval' can
            be passed to evaluate the movement based on data for a specific
            interval.
        """
        if "interval" not in kwargs:
            kwargs["composite"] = True
        prices = self.prices.get(lose_single_symbol=True, fill="both", **kwargs)
        if isinstance(prices.index, pd.IntervalIndex):
            prices = prices.pt.indexed_left
        df = max_decline(prices, label=self.symbol)
        if style:
            caption = f"Maximum Decline {period_string(**kwargs)}"
            return style_df(df, caption=caption)
        else:
            return df

    def corr(
        self, other: "Analysis", method: str = "pearson", style: bool = True, **kwargs
    ) -> float | pd.Styler:
        """Correlation with another Analysis object.

        Correlation will be based on most granular constant interval
        avaiable for period. Pass 'interval' within kwargs to force
        corrleation with data for a specific interval.

        Parameters
        ----------
        method
            As pd.DataFrame.corr() 'method' option.

        **kwargs
            Parameters to define period / price data to be analysed. See
            method doc with `help(analysis.__doc__)`. Cannot include
            'composite' or 'lose_single_symbol'.
        """
        # corr acts on Series and takes a Series as other
        subset = self.prices.get(close_only=True, **kwargs).iloc[:, 0]
        other_subset = other.prices.get(close_only=True, **kwargs).iloc[:, 0]
        corr = subset.corr(other_subset, method=method)
        if not style:
            return corr
        df = pd.DataFrame([corr], index=[self.symbol], columns=[other.symbol])
        caption = (
            f"Correlation {period_string(**kwargs)} "
            + f" ({range_string(subset.index, shand=True)})"
        )
        return style_df_grid(
            df,
            format="{:.2f}",
            caption=caption,
            gradient_kwargs={"cmap": "RdBu", "vmin": -1, "vmax": 1},
        )

    def movements(
        self,
        interval: mp.intervals.RowInterval,
        trend_kwargs: dict,
        trend_cls: type[TrendsProto] = Trends,
        **kwargs,
    ) -> MovementsSupportChartAnaly:
        """Evaluate trends over a given period.

        Parameters
        ----------
        interval
            Price data interval to use for analysis, as 'interval`
            parameter described by `help(self.prices.get)`.

        trend_kwargs
            Kwargs to pass to the `trend_cls`. Do not include 'data' or
            'interval'.

            For the default `trend_cls` (`trends.Trends`) the kwargs are
            'prd', 'ext_break', 'ext_limit' and 'min_bars'. See
            documentation for trends class with
            `help(trends.Trends.__doc__)`.

        trend_cls
            Class to use to evaluate trends. Must fulfil protocol described
            by `trends_base.TrendsProto`.

            Default is `trends.Trends`.

        **kwargs:
            Parameters to define period / price data to be analysed. See
            module doc with `help(analysis.__doc__)`.
        """
        data = self.prices.get(interval, lose_single_symbol=True, **kwargs)
        if isinstance(data.index, pd.IntervalIndex):
            data.index = data.index.left
        return trend_cls(data, interval, **trend_kwargs).get_movements()

    def trends_chart(
        self,
        interval: mp.intervals.RowInterval,
        trend_kwargs: dict,
        gui_cls: type[TrendsGuiBase] = TrendsGui,
        max_ticks: int | None = None,
        log_scale: bool = True,
        display: bool = True,
        **kwargs,
    ) -> TrendsGuiBase:
        """Visualise trends on an OHLC chart.

        Underlying trends data can be accessed via the following attributes
        of the returned object:
            cases: Movements

            trends: Instance of trends class responsible for evaluating
            movement.

        Parameters
        ----------
        interval
            Price data interval to use for analysis, as 'interval`
            parameter described by `help(self.prices.get)`.

        trend_kwargs
            Kwargs to pass to the `gui_cls`. Do not include 'analysis' or
            'interval'.

            For the default `gui_cls` (`TrendsGui`) the kwargs are
            'prd', 'ext_break', 'ext_limit' and 'min_bars'. See
            documentation for associated trends class with
            `help(trends.Trends.__doc__)`.

        gui_cls
            Class to use to create gui to visualise evaluated trends. Must
            be a subclass of `TrendsGuiBase`.

            Default is `TrendsGui`.

        max_ticks
            Maximum number of bars (x-axis ticks) that will shown by
            default (more can be shown via slider). None for no limit.

        log_scale
            True to plot prices against a log scale. False to plot prices
            against a linear scale.

        display
            False to not display the GUI.

        **kwargs:
            Parameters to define period / price data to be analysed. See
            module doc with `help(analysis.__doc__)`.
        """
        return gui_cls(
            self,
            interval,
            **trend_kwargs,
            max_ticks=max_ticks,
            log_scale=log_scale,
            display=display,
            **kwargs,
        )


class Compare(Base):
    """Analyse and compare multiple instruments.

    Properties
    ----------
    symbols -> list[str]:
        Symbols of all instruments being analysed.

    Methods
    -------
    analysis() -> Analysis:
        Analysis object for specific symbol.

    plot():
        Chart rebased prices over a specified period.

    max_adv() -> pd.DataFrame | Styler:
        Maximum percentage advance over specified period.

    max_dec() -> pd.DataFrame | Styler:
        Maximum percentage decline over specified period.

    relative_strength() -> pd.DataFrame | Styler:
        Relative strength between specified instruments over specified period.

    relative_strength_max_adv() -> pd.DataFrame | Styler:
        Relative strength of symbols maximum advances over specified period.

    relative_strength_max_dec() -> pd.DataFrame | Styler:
        Relative strength of symbols maximum declines over specified period.

    corr() -> pd.DataFrame | Styler:
        Correlation between specified instruments over specified period.

    Static Methods
    --------------
    max_chg_compare() -> pd.DataFrame | Styler:
        Maximum change, over specific price data, for multiple instruments.
    """

    PctChgBarCls = guis.GuiPctChgMult

    @staticmethod
    def max_chg_compare(
        direction: Literal["advance", "decline"],
        prices: pd.DataFrame,
        style: bool = True,
        **kwargs,
    ) -> pd.DataFrame | Styler:
        """Maximum change of multiple instruments.

        Parameters
        ----------
        direction
            'advance' for maximum advance, 'decline' for maximum decline.

        prices
            Price table from which to evaluate maximum change.

        style
            True to format ouput and return as Styler.

        **kwargs
            Parameters that were passed to define `prices`.
        """
        func = max_advance if direction == "advance" else max_decline
        dfs = []
        for symbol, s_prices in prices.T.groupby(level="symbol"):
            s_prices = s_prices.T
            s_prices.columns = s_prices.columns.droplevel(level=0)
            if isinstance(s_prices.index, pd.IntervalIndex):
                s_prices = s_prices.pt.indexed_left
            dfs.append(func(s_prices, label=symbol))
        df = pd.concat(dfs)

        dur_cols = [c for c in DURATION_COLUMNS if c in df]
        for col in dur_cols:
            if col in df:
                df.fillna({col: 0}, inplace=True)
            if col == "days":
                df[col] = df[col].astype("int64")

        if direction == "advance":
            df = df[["start", "low", "end", "high", "pct_chg"] + dur_cols]
        else:
            df = df[["start", "high", "end", "low", "pct_chg"] + dur_cols]
        if not style:
            return df
        if len(df.index) > 1:
            av_c = None if not dur_cols else dur_cols[0]
            sum_cols = ["pct_chg"] if av_c is None else ["pct_chg", av_c]
            df = add_summary_row(df, ("mean", sum_cols), label="Av.")
        direction_ = "Advance" if direction == "advance" else "Decline"
        caption = f"Maximum {direction_} {period_string(**kwargs)}"
        return style_df(df, caption=caption)

    def __init__(self, prices: mp.PricesBase):
        """Construct `Compare` instance.

        Parameters
        ----------
        prices:
            Instance of a subclass of `market_prices.PricesBase` from
            which to request price data for the multiple symbols to be
            compared. For example:
                prices = market_prices.PricesYahoo("MSFT, AZN.L")

        Raises
        ------
        ValueError
            If `prices` gets prices for only a single symbol.
        """
        if len(prices.symbols) == 1:
            msg = (
                "The Compare class requires a `prices` instance that gets"
                "price data for multiple symbols, although the past instance"
                f"gets prices for only one: {prices.symbols[0]}."
            )
            raise ValueError(msg)
        super().__init__(prices)

    @property
    def symbols(self) -> list[str]:
        return self.prices.symbols

    def analysis(self, symbol: str) -> Analysis:
        """Analysis object for a given symbol."""
        prices = self.prices.prices_for_symbols(symbol)
        return Analysis(prices)

    def relative_strength(self, style: bool = True, **kwargs) -> pd.DataFrame | Styler:
        """Relative strength over specified period.

        Parameters
        ----------
        **kwargs
            Parameters to define period / price data to be analysed. See
            method doc with `help(analysis.__doc__)`. Cannot include
            'composite' or 'lose_single_symbol'.
        """
        subset = self.prices.get(fill="both", add_a_row=True, close_only=True, **kwargs)
        pct_chg = pct_chg_top_to_bot(subset)
        df = pd.DataFrame({s: pct_chg - v for s, v in pct_chg.items()})
        df.columns.name = "symbol"
        if not style:
            return df
        if len(df.index) > 1:
            df = add_summary(df, "mean", axis="column", label="Av.")
        caption = (
            f"Relative Strength {period_string(**kwargs)} "
            + f"({range_string(subset.index, close=True, shand=True)})"
        )
        return style_df_grid(df, caption=caption)

    def plot(
        self,
        interval: mp.intervals.RowInterval | None = None,
        engine: ChartEngine = "bqplot",
        rebase_on_zoom: bool = True,
        max_ticks: int | None = None,
        log_scale: bool = True,
        **kwargs,
    ) -> guis.GuiMultLine | mpl.artist.Artist:
        """Chart rebased close prices over specified period.

        Parameters
        ----------
        interval
            Price data interval to use, as 'interval` parameter described
            by `help(self.prices.get)`.

            Default: None if `engine` "bqplot" otherwise '1d'.

        engine: Literal["bqplot", "matplotlib"]
            Chart backend

        rebase_on_zoom
            True to rebase prices following zoom. Only implemented if
            `engine` is "bqplot".

        max_ticks
            Maximum number of x-axis ticks that will shown by default (more
            can be shown via slider). None for no limit. Only implemented
            if `engine` is "bqplot".

        log_scale
            True for price axis to be to log scale. Only implemented for
            if `engine` is "bqplot".

        **kwargs
            Parameters to define period / price data to be analysed. See
            method doc with `help(analysis.__doc__)`. Cannot include
            'composite' or 'lose_single_symbol'.
        """
        if engine == "bqplot":
            return guis.GuiMultLine(
                self, interval, rebase_on_zoom, max_ticks, log_scale, **kwargs
            )
        else:
            interval = "1d" if interval is None else interval
            subset = self.prices.get(close_only=True, **kwargs)
            return rebase_to_row(subset).plot()

    def _max_chg(self, direction: Literal["advance", "decline"], style=True, **kwargs):
        """Maximum change over specified period."""
        if "interval" not in kwargs:
            kwargs["composite"] = True
        prices = self.prices.get(fill="both", **kwargs)
        return self.max_chg_compare(direction, prices, style, **kwargs)

    def max_adv(self, style: bool = True, **kwargs) -> pd.DataFrame | Styler:
        """Maximum percentage advance over specified period.

        Parameters
        ----------
        **kwargs
            Parameters to define period / price data to be analysed. See
            method doc with `help(analysis.__doc__)`.

            By default return will be evaluted from data of the highest
            interval that can most accurately reflect the period start and
            end (either as passed or evaluated). This will be a composite
            table if the 'end' could be represented more accurately by a
            lower interval for which data is not otherwise available to
            cover the full requested period. Alternatively, 'interval' can
            be passed to evaluate the movement based on data for a specific
            interval.
        """
        return self._max_chg("advance", style, **kwargs)

    def max_dec(self, style: bool = True, **kwargs) -> pd.DataFrame | Styler:
        """Maximum percentage decline over specified period.

        Parameters
        ----------
        **kwargs
            Parameters to define period / price data to be analysed. See
            method doc with `help(analysis.__doc__)`.

            By default return will be evaluted from data of the highest
            interval that can most accurately reflect the period start and
            end (either as passed or evaluated). This will be a composite
            table if the 'end' could be represented more accurately by a
            lower interval for which data is not otherwise available to
            cover the full requested period. Alternatively, 'interval' can
            be passed to evaluate the movement based on data for a specific
            interval.
        """
        return self._max_chg("decline", style, **kwargs)

    def _relative_strength_max_chg(
        self, func: Callable, style: bool = True, **kwargs
    ) -> pd.DataFrame | Styler:
        """Relative strength of maximum change over specified period."""
        df = func(style=False, **kwargs)
        srs = df["pct_chg"]
        rel_str_df = pd.DataFrame({s: srs - v for s, v in srs.items()})
        if not style:
            return rel_str_df
        if len(rel_str_df.index) > 1:
            rel_str_df = add_summary(rel_str_df, "mean", axis="both", label="Av.")
        styler = func(style=True, **kwargs)
        caption = f"Relative Strength {styler.caption}"
        return style_df_grid(rel_str_df, caption=caption)

    def relative_strength_max_adv(
        self, style: bool = True, **kwargs
    ) -> pd.DataFrame | Styler:
        """Relative strength of maximum advance over specified period.

        Parameters as for `max_adv`.
        """
        return self._relative_strength_max_chg(self.max_adv, style=style, **kwargs)

    def relative_strength_max_dec(
        self, style: bool = True, **kwargs
    ) -> pd.DataFrame | Styler:
        """Relative strength of maximum decline over specified period.

        Parameters as for `max_dec`.
        """
        return self._relative_strength_max_chg(self.max_dec, style=style, **kwargs)

    def corr(
        self, method: str = "pearson", style: bool = True, **kwargs
    ) -> pd.DataFrame | Styler:
        """Correlation over specified period.

        Correlation will be based on LEAST granular constant interval
        that can most accurately represent the period start and end. Pass
        `interval` within kwargs to force corrleation to be against a
        specific interval.

        Parameters
        ----------
        method
            As pd.DataFrame.corr() 'method' option.

        **kwargs
            Parameters to define period / price data to be analysed. See
            method doc with `help(analysis.__doc__)`. Cannot include
            'close_only'.
        """
        subset = self.prices.get(fill="both", close_only=True, **kwargs)
        df = subset.corr(method=method)
        if not style:
            return df
        if len(df.index) > 1:
            df = add_summary(df, "sum", axis="column", label="Av.")
        df["Av."] = (df["Av."] - 1) / (len(df.columns) - 2)
        caption = (
            f"Correlation {period_string(**kwargs)}" f" ({range_string(subset.index)})"
        )
        return style_df_grid(
            df,
            format="{:.2f}",
            gradient_kwargs={"cmap": "RdBu", "vmin": -1},
            caption=caption,
        )



"""Utilities for operating on pandas DataFrame and Series."""

from typing import Any, Literal
from zoneinfo import ZoneInfo

import pandas as pd
import numpy as np


def rebase_to_cell(
    data: pd.DataFrame | pd.Series,
    row: Any,
    col: str,
    value: int = 100,
) -> pd.DataFrame:
    """Rebase a pandas object to a given base value for a given cell.

    Parameters
    ----------
    data
        pandas object to rebase.

    row
        Label of row containing cell to have base value.

    col
        Label of column containing cell to have base value.

    value
        Base value. Cell at `row`, `column` will have this value with all
        other cells rebased relative to this base value.

    Example
    >>> df = pd.DataFrame(dict(open_=range(5), close_=range(5, 10)))
    >>> df
       open_  close_
    0      0       5
    1      1       6
    2      2       7
    3      3       8
    4      4       9
    >>> rebase_to_cell(df, 3, 'close_', 100)
       open_  close_
    0    0.0    62.5
    1   12.5    75.0
    2   25.0    87.5
    3   37.5   100.0
    4   50.0   112.5
    """
    return data / data.loc[row, col] * value


def rebase_to_row(
    data: pd.DataFrame | pd.Series, row: int = 0, value: int = 100
) -> pd.DataFrame:
    """Rebase a pandas object to a given row.

    If `data` is a pd.DataFrame, each column will be rebased independently.

    Parameters
    ----------
    data
        pandas object to rebase.

    row
        index of row against which to rebase data.

    value
        Base value for each cell of `row`. All other rows will be rebased
        relative to this base value.

    Examples
    --------
    >>> df = pd.DataFrame(dict(open_=range(5), close_=range(5, 10)))
    >>> df
       open_  close_
    0      0       5
    1      1       6
    2      2       7
    3      3       8
    4      4       9
    >>> rebase_to_row(df, 2, 100)
       open_      close_
    0    0.0   71.428571
    1   50.0   85.714286
    2  100.0  100.000000
    3  150.0  114.285714
    4  200.0  128.571429
    """
    return data / data.iloc[row] * value


def tolists(df: pd.DataFrame) -> list[list[Any]]:
    """Convert pd.DataFrame to list of lists.

    Each each inner list represents one column of the DataFrame.

    Parameters
    ----------
    df
        DataFrame to convert to list of lists.

    Examples
    --------
    >>> df = pd.DataFrame(dict(open_=range(5), close_=range(5, 10)))
    >>> df
       open_  close_
    0      0       5
    1      1       6
    2      2       7
    3      3       8
    4      4       9
    >>> tolists(df)
    [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]
    """
    return [col.values.tolist() for name, col in df.items()]


def interval_of_intervals(
    intervals: pd.IntervalIndex,
    closed: Literal["left", "right", "both", "neither"] = "right",
) -> pd.Interval:
    """Return interval covered by a monotonic IntervalIndex.

    Parameters
    ----------
    intervals
        Monotonic pd.IntervalIndex for which require the encompassing
        interval.

    closed
        Side on which to close returned pd.Interval.

    Raises
    ------
    ValueError
        If `intervals` is not monotonic.

    Examples
    --------
    >>> # ignore first part, for testing purposes only...
    >>> import pytest, pandas
    >>> v = pandas.__version__
    >>> if (
    ...     (v.count(".") == 1 and float(v) < 2.2)
    ...     or (
    ...         v.count(".") > 1
    ...         and float(v[:v.index(".", v.index(".") + 1)]) < 2.2
    ...     )
    ... ):
    ...     pytest.skip("printed return only valid from pandas 2.2")
    >>> #
    >>> # example from here...
    >>> left = pd.date_range('2021-05-01 12:00', periods=5, freq='h')
    >>> right = left + pd.Timedelta(30, 'min')
    >>> index = pd.IntervalIndex.from_arrays(left, right)
    >>> index.to_series(index=range(5))
    0    (2021-05-01 12:00:00, 2021-05-01 12:30:00]
    1    (2021-05-01 13:00:00, 2021-05-01 13:30:00]
    2    (2021-05-01 14:00:00, 2021-05-01 14:30:00]
    3    (2021-05-01 15:00:00, 2021-05-01 15:30:00]
    4    (2021-05-01 16:00:00, 2021-05-01 16:30:00]
    dtype: interval
    >>> interval_of_intervals(index)
    Interval(2021-05-01 12:00:00, 2021-05-01 16:30:00, closed='right')
    """
    # NOTE Can lose doctest skip when pandas min support is >= 2.2
    if not intervals.is_monotonic_increasing:
        raise ValueError(f"`intervals` must be monotonic. Received as '{intervals}'.")
    return pd.Interval(intervals[0].left, intervals[-1].right, closed=closed)


def interval_contains(interval: pd.Interval, intervals: pd.IntervalIndex) -> np.ndarray:
    """Boolean array indicating intervals contained within an interval.

    Parameters
    ----------
    interval: pd.Interval
        Intervals within which to check if intervals are contains.

    intervals: pd.IntervalIndex
        Intervals to check if contained in interval.
    """
    if interval.closed == "left":
        left_cond = intervals.left >= interval.left
        right_cond = intervals.right < interval.right
    elif interval.closed == "right":
        left_cond = intervals.left > interval.left
        right_cond = intervals.right <= interval.right
    elif interval.closed == "both":
        left_cond = intervals.left >= interval.left
        right_cond = intervals.right <= interval.right
    else:
        left_cond = intervals.left > interval.left
        right_cond = intervals.right < interval.right
    return left_cond & right_cond


def intervals_subset(
    intervals: pd.IntervalIndex, interval: pd.Interval, overlap: bool = False
) -> pd.IntervalIndex:
    """Subset of intervals overlapping with or contained by an interval.

    Parameters
    ----------
    intervals
        Intervals from which to take subset.

    interval
        Interval of intervals to be taken as a subset.

    overlap
        True to include all intervals that overlap with interval, False to
        only return intervals that are fully contained within interval.
    """
    if overlap:
        bv = intervals.overlaps(interval)
    else:
        bv = interval_contains(interval, intervals)
    return intervals[bv]


def interval_index_new_tz(
    index: pd.IntervalIndex, tz: ZoneInfo | str | None
) -> pd.IntervalIndex:
    """Return pd.IntervalIndex with different timezone.

    Note: `index` is not changed in place.

    Parameters
    ----------
    index
        pd.IntervalIndex on which to base return. Must have left and right
        sides as pd.DatetimeIndex. If these pd.DatetimeIndex are tz-naive
        then indices will be localised to `tz`, if tz_aware then indices
        will be converted to `tz` if `tz` is not None, or localized to
        None if `tz` is None.

    tz
        Timezone for returned pd.IntervalIndex. Examples: "US/Eastern",
        "Europe/Paris", "UTC".

        Pass as None to return as timezone naive.

    Returns
    -------
    pd.IntervalIndex
        pd.IntervalIndex as `index` albeit with timezone as `tz`.

    Examples
    --------
    >>> tz = ZoneInfo("US/Central")
    >>> left = pd.date_range(
    ...     '2021-05-01 12:00', periods=5, freq='h', tz=tz
    ... )
    >>> right = left + pd.Timedelta(30, 'min')
    >>> index = pd.IntervalIndex.from_arrays(left, right)
    >>> index.right.tz
    zoneinfo.ZoneInfo(key='US/Central')
    >>> UTC = ZoneInfo("UTC")
    >>> new_index = interval_index_new_tz(index, tz=UTC)
    >>> new_index.left.tz == new_index.right.tz == UTC
    True
    >>> tz_naive = interval_index_new_tz(index, tz=None)
    >>> tz_naive[0].left
    Timestamp('2021-05-01 12:00:00')
    """
    indices = []
    for indx in [index.left, index.right]:
        try:
            indices.append(indx.tz_localize(tz))
        except TypeError:
            indices.append(indx.tz_convert(tz))
    return pd.IntervalIndex.from_arrays(indices[0], indices[1], closed=index.closed)


def index_dates_to_str(index: pd.DatetimeIndex) -> pd.Index:
    """Convert index representing dates to an index of dtype 'string'.

    Formats dates as %Y-%m-%d.

    Examples
    --------
    >>> import pandas as pd
    >>> index = pd.date_range("2020-01-09", "2020-01-10", freq="D")
    >>> str_index = index_dates_to_str(index)
    >>> str_index
    Index(['2020-01-09', '2020-01-10'], dtype='string')
    """
    return index.strftime("%Y-%m-%d").astype("string")



"""Definitions of widgets to contribute to GUI.

Widgets define the GUI, they do NOT define any event handling.

Functions
---------
loading_overlay() -> v.Overlay:
    Overlay application with transparent sheet with loading symbol.

close_but() -> wu.ButtonIcon:
    Return 'Close' icon button.

legend_but() -> vu.IconButton:
    Return `IconButton` to cycle legend location.

rebase_but() -> vu.IconBut:
    Return `IconButton` to rebase a multiline chart.

Classes
-------
IntervalSelector:
    ToggleButtons to select a date interval.

IconRowTop:
    Row of icons to place above chart with selectable dates.

TabsControl:
    Tabs for price chart operations.

HtmlOutput:
    HTML output box.

BarDirectionToggle:
    Toggle to set bar chart direction.

BarTypeToggle:
    Toggle to set bar chart type.

PctChgIconRow:
    Percent change bar chart options for single instrument.

PctChgIconRowMult:
    Percent change bar chart options for multiple instruments.
"""

from collections.abc import Callable
import re

import ipywidgets as w
import ipyvuetify as v
from market_prices import intervals

import market_analy.utils.ipywidgets_utils as wu
import market_analy.utils.ipyvuetify_utils as vu
from market_analy.utils.dict_utils import set_kwargs_from_dflt

ICON_DIM = "35px"

TT_KWARGS_DFLT = {"bottom": True, "open_delay": 750, "content_class": "black--text"}

BG = "grey darken-4"


def close_but(handler: Callable | None = None) -> wu.ButtonIcon:
    """Return close icon button. Red with white cross.

    Parameters
    ----------
    handler
        On-click handler. Signature should accommodate following parameters:
            handler_name(widget)
    """
    layout = w.Layout(align_self="flex-start", width=ICON_DIM)
    but = wu.ButtonIcon.close_(layout=layout)
    if handler is not None:
        but.on_click(handler)
    but.tooltip = "Close"
    return but


class IntervalSelector(wu.ToggleButtonsHandled):
    """`w.ToggleButtons` to select a date interval."""

    def __init__(
        self,
        labels: list[str],
        handler: Callable | None = None,
        value: intervals.PTInterval | None = None,
        layout_kwargs: dict | None = None,
    ):
        """Constructor.

        Parameters
        ----------
        labels
            Button labels. Should be `market-prices` shorthand intervals
            from which PTIntervals can be instantiated, for example:
                ['3mo', '1d', '1h', '30m', '5m']

        handler
            Handler to handle changes to `ToggleButtons` value. Should have
            signature `handler(change: dict)` where 'change' is a
            dictionary including keys 'old' and 'new' which have values as
            old and new price intervals, of type `intervals.PTInterval`.

        value
            Value corresponding with any toggle button to initially select.
            If not passed then first toggle button will be selected.

        layout_kwargs
            Passed to `w.Layout` assigned to `ToggleButtons`.
        """
        values = [intervals.to_ptinterval(lab) for lab in labels]
        tooltips = ["Select tick interval" for _ in range(len(values))]
        super().__init__(
            labels,
            values,
            tooltips=tooltips,
            handler=handler,
            value=value,
            layout_kwargs=layout_kwargs,
        )


class DrawdownSelector(wu.ToggleButtonsHandled):
    """`w.ToggleButtons` to select a drawdown period.

    Button values generally represent the value that can be passed
    through to the 'window' option of pd.DataFrame.rolling. Values are:
        - as `labels` for those which represent perioods defined in bars.
        - '365D' for '52wk' button
        - 'ATH' for 'ATH' button (cannot take None as this intefers with
            the buttons implemnentation)
        - False for 'x' button

    Parameters
    ----------
    labels
        Button labels representing the options that can be selected to
        determine the period, in number of bars, over which the high is
        evaluated. NB 'ATH' and '52wk' will always be added to the start
        of the options by the constructor.

    handler
        Handler to handle changes to `ToggleButtons` value. Should have
        signature `handler(change: dict)` where 'change' is a dictionary
        including keys 'old' and 'new' which have values as old and new
        values.

    value
        Value corresponding with any toggle button to initially select. If
        not passed then first toggle button will be selected.
    """

    def __init__(
        self,
        labels: list[int],
        handler: Callable | None = None,
        value: str | int | None = None,
    ):
        labels_ = ["ATH", "52wk"] + [str(lbl) for lbl in labels] + ["x"]
        values = ["ATH", "365D"] + [lbl for lbl in labels] + [False]
        tooltips = (
            ["All-time high", "52 week high"]
            + [f"{lbl} bar high" for lbl in labels]
            + ["Hide drawdown"]
        )
        super().__init__(
            labels_,
            values,
            tooltips=tooltips,
            handler=handler,
            value=value,
        )


def loading_overlay() -> v.Overlay:
    """Overlay application with opaque sheet with loading symbol."""
    prog = v.ProgressCircular(indeterminate=True, color="white", size=30)
    return v.Overlay(color="grey", value=False, absolute=True, children=[prog])


class Dialog(vu.Dialog):
    """Dialog box.

    Pass instance as child of component to be overlaid and within which
    dialog box to appear.
    """

    def __init__(self, title: str | None = None, text: str | None = None):
        title = title if title is not None else "title placeholder"
        text = text if text is not None else "text placeholder"
        super().__init__(title, text)


class IconRowTop(wu.IconRow):
    """Row of icons to place above chart with selectable dates.

    Includes following icons:
        'max_dates'
        'refresh' (only if 4 handlers passed to constructor)
        'resize_window'
        'close'

    Constructor accommodates on-click handlers.

    Parameters
    ----------
    handlers
        3 or 4-list of callables to serve as on-click handlers for,
        respectively, `IconButtons` 'max_dates', 'refresh' (only if 4)
        'resize_window' and 'close'.
    """

    def __init__(self, handlers: list[Callable] | None = None, **kwargs):
        assert handlers is None or len(handlers) in [3, 4]
        has_3_handlers = handlers is not None and len(handlers) == 3
        icon_button_names = ["max_dates", "refresh", "resize_window", "close_"]
        if has_3_handlers:
            icon_button_names.remove("refresh")

        super().__init__(icon_button_names, handlers=handlers, **kwargs)

        tooltips = ["Max dates", "Reset", "Resize Chart", "Close"]
        if has_3_handlers:
            tooltips.remove("Reset")
        for i, child in enumerate(self.children):
            child.tooltip = tooltips[i]


class TabsControl(v.Tabs):
    """Tabs widget.

    Includes:
        Cursor tab, one row:
            ToggleIcons[+ - IconBut] // Divider // Lightbulb IconBut
                // Trash IconBut
        Selector tab, one row:
            Zoom IconBut // UpArrow IconBut // DownArrow IconBut

    Attributes
    ----------
    cursor_objs: `vue.Layout`
        Container of widgets that comprise cursor tab items.
    cursor_toggle: `ToggleIcons`
    but_lightbulb: `IconBut`
    self.but_trash: `IconBut`

    slctr_objs: `vue.Layout`
        Container of widgets that comprise selector tab items.
    but_zoom: `IconBut`
    but_arrow_up: `IconBut`
    but_arrow_down: `IconBut`
    """

    _tab_container_class_ = "d-flex justify-center align-center"

    def __init__(self):
        self.cursor_objs: v.Layout
        self.cursor_toggle: vu.ToggleIcons
        self.but_lightbulb: vu.IconBut
        self.but_trash: vu.IconBut
        self._create_cursor_tab_content()

        self.slctr_objs: v.Layout
        self.but_zoom: vu.IconBut
        self.but_arrow_up: vu.IconBut
        self.but_arrow_down: vu.IconBut
        self._create_slctr_tab_content()

        tab_class_ = "text-lowercase font-weight-regular " + BG

        texts = ["Cursor", "Selector"]
        tooltips = ["Crosshair Operations", "Operate on a selected period"]
        tts_kwargs = [{"left": True}, {"right": True}]
        for kwargs in tts_kwargs:
            kwargs["open_delay"] = TT_KWARGS_DFLT["open_delay"]
        tab_tt_list = []
        for text, tt, tt_kwargs in zip(texts, tooltips, tts_kwargs):
            tab_tt = vu.TabTt(
                children=[text], class_=tab_class_, tooltip=tt, tt_kwargs=tt_kwargs
            ).tt
            tab_tt_list.append(tab_tt)

        tab_items = [
            v.TabItem(children=[layout], class_=BG)
            for layout in [self.cursor_objs, self.slctr_objs]
        ]

        tabs_class_ = "d-flex flex-column align-center flex-grow-0 " + BG
        super().__init__(
            v_model=0,
            children=tab_tt_list + tab_items,
            centered=True,
            fixed_tabs=True,
            dark=True,
            slider_size=1,
            height="30px",
            class_=tabs_class_,
        )

    def _create_cursor_tab_content(self):
        colors = ["green", "red"]
        d = {
            "icon_names": ["fa-plus", "fa-minus"],
            "names": ["plus", "minus"],
            "tooltips": ["Add crosshair", "Remove crosshair"],
            "deselected_colors": ["green darken-4", "red darken-4"],
            "selected_colors": colors,
        }

        d["tts_kwargs"] = [
            set_kwargs_from_dflt({"color": c}, TT_KWARGS_DFLT) for c in colors
        ]

        self.cursor_toggle = vu.ToggleIcons(**d)

        divider = vu.IconBut.divider()

        tt = "Toggle crosshairs transparency"
        color = "yellow"
        tt_kwargs = {"color": color}
        set_kwargs_from_dflt(tt_kwargs, TT_KWARGS_DFLT)
        self.but_lightbulb = vu.ToggleIcon(
            "fa-lightbulb-o",
            selected=True,
            deselected_color=color + " darken-4",
            selected_color=color,
            tooltip=tt,
            tt_kwargs=tt_kwargs,
        )

        tt = "Remove all crosshairs"
        color = "red accent-3"
        tt_kwargs = {"color": color}
        set_kwargs_from_dflt(tt_kwargs, TT_KWARGS_DFLT)
        self.but_trash = vu.IconBut(
            "fa-trash", color=color, tooltip=tt, tt_kwargs=tt_kwargs
        )

        cursor_buts = [self.but_lightbulb.tt, self.but_trash.tt]
        children = self.cursor_toggle.contain_me + [divider] + cursor_buts
        self.cursor_objs = v.Layout(
            children=children, class_=self._tab_container_class_
        )

    def _create_slctr_tab_content(self):
        def get_but(icon_name: str, color: str, tooltip: str) -> vu.IconBut:
            class_ = "mx-4 mb-2 mt-4"
            tt_kwargs = {"color": color}
            set_kwargs_from_dflt(tt_kwargs, TT_KWARGS_DFLT)
            return vu.IconBut(
                icon_name,
                color=color,
                class_=class_,
                tooltip=tooltip,
                tt_kwargs=tt_kwargs,
            )

        self.but_zoom = get_but(
            "fa-search-plus", "orange lighten-2", "Zoom to selection"
        )

        tt = "Evalute maximum advance over selected period"
        self.but_arrow_up = get_but("fa-arrow-circle-up", "light-green lighten-1", tt)

        tt = "Evalute maximum decline over selected period"
        self.but_arrow_down = get_but("fa-arrow-circle-down", "red lighten-1", tt)

        buts = [self.but_zoom.tt, self.but_arrow_up.tt, self.but_arrow_down.tt]
        self.slctr_objs = v.Layout(children=buts, class_=self._tab_container_class_)

    def reset(self):
        self.v_model = 0
        self.cursor_toggle.deselect()


class HtmlOutput(w.HBox):
    """HTML Output box with button to close display."""

    def __init__(self):
        self._html = w.HTML()

        self._close_but = wu.ButtonIcon.close_(
            tooltip="Close output", layout=w.Layout(height="30px")
        )
        self._close_but.on_click(self._close_button_handler)
        self._hide_close_button()

        super().__init__(
            children=[self._html, self._close_but],
            layout=w.Layout(
                justify_content="center",
                align_self="center",
                grid_gap="10px",
                margin="20px 0 0 0",
            ),
        )

    def _show_close_button(self):
        self._close_but.layout.visibility = "visible"
        self._close_but.layout.margin = "0 0 0 0"  # "25px 0 0 0"

    def _hide_close_button(self):
        self._close_but.layout.visibility = "hidden"
        self._close_but.layout.margin = "0 0 0 0"

    def display(self, html: str, add_padding: int = 0):
        """Display html.

        Parameters
        ----------
        html
            HTML to be displayed.

        add_padding
            Where html includes a table, add padding in the form of blank
            spaces to the start of the contents of every 'td' and 'tr'
            tabs. Pass `add_padding` as number of spaces to be included.

        Notes
        -----
        add_padding option provided to account for padding style
        information being annoyingly lost, or at least not recognised,
        when a `HtmlOutput` widget is otherwise included to a
        `ipyvuetify.app`.
        """
        if add_padding:

            def repl(mo: re.Match) -> str:
                match = mo.group()
                return f'>{"&nbsp" * add_padding}{match[1:]}'

            regex = re.compile(r"\>.+(?=.*\</t[hd]\>)")
            html = re.sub(regex, repl, html)

        self._html.value = html
        self._show_close_button()

    def clear(self):
        """Clear output."""
        self._html.value = ""
        self._hide_close_button()

    def _close_button_handler(self, widget):
        self.clear()


class _BiToggle(vu.ToggleIcons):
    """Two `IconButtons`, one of which has to be selected."""

    def __init__(
        self,
        icon_names: list[str],
        names: list[str],
        color: str,
        tooltips: list[str],
        **kwargs,
    ):
        d = {
            "icon_names": icon_names,
            "names": names,
            "deselected_colors": ["grey lighten-1"] * 2,
            "selected_colors": [color] * 2,
            "tooltips": tooltips,
            "tts_kwargs": [
                {
                    "bottom": True,
                    "color": color,
                    "open_delay": 750,
                    "content_class": "white--text",
                }
            ]
            * 2,
            "select_by": "name",
            "allow_no_selection": False,
            "initial_selection": names[0],
        }
        super().__init__(**d, **kwargs)


class BarDirectionToggle(_BiToggle):
    """Toggle to set bar chart direction.

    Two `IconButton` toggle to set bar chart direction, vertical or
    horizontal.
    """

    def __init__(self):
        d = {
            "icon_names": ["fa-grip-lines-vertical", "fa-grip-lines"],
            "names": ["vertical", "horizontal"],
            "color": "orange",
            "tooltips": ["vertical bars", "horizontal bars"],
        }
        super().__init__(**d)


class BarTypeToggle(_BiToggle):
    """Toggle to set bar chart type.

    Two `IconButton` toggle to set bar chart type, stacked or
    grouped.
    """

    def __init__(self):
        d = {
            "icon_names": ["mdi-chart-bar-stacked", "mdi-chart-bar"],
            "names": ["stacked", "grouped"],
            "color": "blue",
            "tooltips": ["stacked", "grouped"],
        }
        super().__init__(**d)


def create_icon_but(
    icon_name: str,
    tooltip: str,
    color: str,
    handler: Callable | None = None,
    **kwargs,
) -> vu.IconBut:
    """`IconButton` with tooltip and handler.

    Parameters
    ----------
    icon_name
        Example, 'fa-list-alt'

    tooltip
        Tooltip text.

    color

    handler
        If passed handler will handle button clicks. Handler should have
        signature:
            handler(widget, event, data)

        If not passed then handler can be subsequently set with:
            but.on_event('click', handler)

    **kwargs
        Passed on to `IconBut`. Will not pass on parameters defined by
        method to maintain legend button style.
    """
    kwargs["icon_name"] = icon_name
    kwargs["tooltip"] = tooltip
    kwargs["color"] = color

    dark = kwargs.get("dark", False) and (dark_color := kwargs.get("dark_color", False))
    tt_kwargs = {"color": dark_color if dark else color}
    tt_kwargs = set_kwargs_from_dflt(tt_kwargs, TT_KWARGS_DFLT)
    tt_kwargs = set_kwargs_from_dflt(kwargs.pop("tt_kwargs", {}), tt_kwargs, deep=True)
    but = vu.IconBut(tt_kwargs=tt_kwargs, **kwargs)
    if handler is not None:
        but.on_event("click", handler)
    return but


def legend_but(handler: Callable | None = None, **kwargs) -> vu.IconBut:
    """`IconButton` to cycle legend location.

    Parameters
    ----------
    handler
        If passed handler will handle clicks on the legend button. Handler
        should have signature:
            handler(widget, event, data)

    **kwargs
        Passed on to `IconBut`. Will not pass on parameters defined by
        method to maintain legend button style.
    """
    return create_icon_but(
        icon_name="fa-list-alt",
        tooltip="Cycle legend location",
        color="white",
        handler=handler,
        **kwargs,
    )


def rebase_but(handler: Callable | None = None, **kwargs) -> vu.IconBut:
    """`IconButton` to rebase a multiline chart.

    Parameters
    ----------
    handler
        If passed handler will handle clicks of rebase button. Handler
        should have signature:
            handler(widget, event, data)

    **kwargs
        Passed on to `IconBut`. Will not pass on parameters defined by
        method to maintain legend button style.
    """
    return create_icon_but(
        icon_name="fa-dot-circle-o",
        tooltip="Rebase prices",
        color="orange",
        handler=handler,
        **kwargs,
    )


class PctChgIconRowMult(v.Layout):
    """Percent change bar chart options for multiple instruments."""

    def __init__(self):
        self.bar_type_tog = BarTypeToggle()
        self.legend_cycle_but = legend_but()

        children = [
            v.Layout(children=self.bar_type_tog.contain_me, class_="flex-grow-0"),
            self.legend_cycle_but.tt,
        ]

        super().__init__(children=children, class_=BG, justify_space_around=True)


class CaseControls(v.Layout):
    """Container of controls for selecting and viewing analysis cases.

    Client should directly assign handlers to exposed buttons after
    creating instance.

    Includes vu.IconBut arranged as:

                        but_next  //  but_prev
        but_show_all //
                        but_narrow  //  but_wide

    Subclasses can extend by extending or overriding the `_create_children`
    method.
    """

    _TURN_OFF_COLOR = "#CD853F"

    def __init__(self):
        self.but_show_all: vu.IconBut
        self.but_next: vu.IconBut
        self.but_prev: vu.IconBut
        self.but_narrow: vu.IconBut
        self.but_wide: vu.IconBut
        self._buts: list[vu.IconBut] = []
        self._dark = True

        super().__init__(
            children=self._create_children(),
            class_="d-flex justify-start ml-2",
        )

    def _create_iconbut(
        self,
        icon_name: str,
        tooltip: str,
        color: str,
        class_="mr-2",
        dark: bool | None = None,
    ) -> vu.IconBut:
        if dark is None:
            dark = self._dark
        but = create_icon_but(
            icon_name=icon_name,
            tooltip=tooltip,
            color=color,
            class_=class_,
            dark_color="grey",
            dark=dark,
        )
        self._buts.append(but)
        return but

    def _create_show_all_container(self) -> v.Layout:
        self.but_show_all = self._create_iconbut(
            icon_name="mdi-chart-scatter-plot",
            tooltip="Toggle visibilty of case marks",
            color="yellow",
            class_="flex-grow-0 align-self-center",
            dark=False,
        )
        return v.Layout(
            children=[
                v.Layout(
                    children=[self.but_show_all.tt],
                    class_="d-flex justify-start flex-grow-0 mr-2",
                )
            ],
            class_="d-flex justify-start flex-grow-0",
        )

    def _create_prev_next_container(self) -> v.Layout:
        self.but_prev = self._create_iconbut(
            "mdi-arrow-left-circle", "Select previous case", "#F0E68C"
        )
        self.but_next = self._create_iconbut(
            "mdi-arrow-right-circle", "Select next case", "#FFD700"
        )
        return v.Layout(
            children=[self.but_prev.tt, self.but_next.tt],
            class_="d-flex justify-start mb-2",
        )

    def _create_narrow_wide_view_container(self) -> v.Layout:
        self.but_narrow = self._create_iconbut(
            "mdi-arrow-expand-horizontal",
            "Narrow view around selected case",
            "#BDB76B",
        )
        self.but_wide = self._create_iconbut(
            "mdi-arrow-left-right", "Wide view around selected case", "#BDB76B"
        )
        return v.Layout(
            children=[self.but_narrow.tt, self.but_wide.tt],
            class_="d-flex justify-start",
        )

    def _create_case_container(self) -> v.Layout:
        contain_prev_next = self._create_prev_next_container()
        contain_narrow_wide = self._create_narrow_wide_view_container()
        return v.Layout(
            children=[contain_prev_next, contain_narrow_wide],
            class_="d-flex flex-column justify-center flex-grow-0 ml-2",
        )

    def _create_children(self) -> list[v.Layout]:
        """Create all children to populate top level container."""
        return [
            self._create_show_all_container(),
            self._create_case_container(),
        ]

    @property
    def is_dark_single_case(self) -> bool:
        """Query if icons for single-case actions are dark."""
        return self._dark

    def darken_single_case(self):
        """Darken icon buttons for single-case actions.

        Will not darken any icon waiting to be switched off.
        """
        for but in self._buts:
            if (but.icon_color == self._TURN_OFF_COLOR) or but is self.but_show_all:
                continue
            but.darken()
        self._dark = True

    def lighten_single_case(self):
        """Lighten icon buttons for single-case actions."""
        for but in self._buts:
            if but is self.but_show_all:
                continue
            but.lighten()
        self._dark = False


class TrendControls(CaseControls):
    """Container of controls for selecting and viewing trend movements.

    Client should directly assign handlers to exposed buttons after
    creating instance.

    Includes vu.IconBut arranged as:

                        but_next  //  but_prev
        but_show_all //                           // but_ruler
                        but_narrow  //  but_wide
    """

    _TURN_OFF_COLOR = "#CD853F"

    def __init__(self):
        self.but_ruler: vu.IconBut

        super().__init__()

    def but_ruler_handler(self, but: vu.IconBut, event: str, data: dict):
        """Handler to toggle ruler but color light/red if group not dark."""
        if but.is_dark:
            return
        if but.is_light:
            but.icon_color = self._TURN_OFF_COLOR
            return
        f = but.darken if self.is_dark_single_case else but.lighten
        f()

    def _create_ruler_container(self) -> v.Layout:
        self.but_ruler = self._create_iconbut(
            icon_name="mdi-ruler",
            tooltip="Show ruler with width of trend period",
            color="yellow",
            class_="flex-grow-0 align-self-center",
        )
        self.but_ruler.on_event("click", self.but_ruler_handler)
        return v.Layout(
            children=[self.but_ruler.tt],
            class_="d-flex justify-start flex-grow-1",
        )

    def _create_children(self) -> list[v.Layout]:
        """Create all children to populate top level container."""
        children = super()._create_children()
        children.append(self._create_ruler_container())
        return children



"""Mathematics utility functions and classes."""

import math


def log10_floor(x: int | float) -> int:
    """Floor of log10."""
    return math.floor(math.log10(abs(x)))


def sigfig(x: int | float, sf: int) -> int | float:
    """Return `x` to `sf` significant figures."""
    int_digits = log10_floor(x) + 1
    places_to_round = int_digits - sf
    return round(x, -places_to_round)



"""GUIs to visualise movements identifeid by Trends classes."""

from __future__ import annotations

import typing
from collections import abc
from collections.abc import Callable

import bqplot as bq
import ipyvuetify as v
import market_prices as mp
import pandas as pd

import market_analy

from ..guis import GuiOHLCCaseBase
from ..gui_parts import TrendControls
from ..utils import ipyvuetify_utils as vu
from ..utils.bq_utils import HFixedRule
from . import charts
from .analy import RvrAlt, Trends, TrendsAlt

if typing.TYPE_CHECKING:
    from .. import analysis as ma_analysis
    from . import TrendsProto
    from .movements import Movement, MovementAlt, Movements, MovementsAlt


class TrendsGuiBase(GuiOHLCCaseBase):
    """GUI to display and interact with trend analysis over OHLC Chart.

    Parameters
    ----------
    analysis
        Analysis instance representing instrument to be plotted.

    interval
        Interval covered by one bar (i.e. one x-axis tick). As 'interval`
        parameter described by `help(analysis.prices.get)`.

    trend_cls
        Class to use to analyse trends. Must conform with
        `trends_base.TrendsProto`, for example, `trends.Trends`.

    trend_kwargs
        Arguments to pass to `trend_cls`. Do not include `data` or
        `interval`.

    max_ticks
        Maximum number of bars (x-axis ticks) that will shown by default
        (client can choose to show more via slider). None for no limit.

    log_scale
        True to plot prices against a log scale. False to plot prices
        against a linear scale.

    display
        True to display created GUI.

    narrow_view
        When displaying a movement in 'narrow' view, the number of bars
        that should be shown before the bar representing the movement's
        start and after the bar representing the movement's confirmed end.

    wide_view
        When displaying a movement in 'wide' view, the number of bars that
        should be shown before the bar representing the movement's start
        and after the bar representing the movement's confirmed end.

    chart_kwargs
        Any kwargs to pass on to the chart class.

    **kwargs
        Period for which to plot prices. Passed as period parameters as
        described by market-prices documentation for 'PricesCls.get'
        method where 'PricesCls' is the class that was passed to
        'PricesCls' parameter of `mkt_anlaysis.Analysis` to intantiate
        `analysis` (for example, documenation for
        'market_prices.PricesYahoo.get').

    Notes
    -----
    -- Subclass Implementation --

    This base class can be subclassed, including for the purposes of:

        Passing through the required `trend_cls` and concreting in its own
        constructor arguments passed on to this base as `trend_kwargs`.

        Concrete constructor arguments such as `narrow_view`, `wide_view`
        etc.
    """

    _HAS_INTERVAL_SELECTOR = False
    _HAS_DRAWDOWN = False

    def __init__(
        self,
        analysis: ma_analysis.Analysis,
        interval: mp.intervals.RowInterval,
        trend_cls: type[TrendsProto],
        trend_kwargs: dict,
        max_ticks: int | None = None,
        log_scale: bool = True,
        display: bool = True,
        narrow_view: int = 10,
        wide_view: int = 10,
        chart_kwargs: dict | None = None,
        **kwargs,
    ):
        data, _, _ = self._set_initial_data(analysis, interval, kwargs)
        if isinstance(data.index, pd.IntervalIndex):
            data.index = data.index.left
            if data.index.tz is not None:
                data.index = data.index.tz_localize(None)
        self.trends = trend_cls(data, interval, **trend_kwargs)
        cases = self.trends.get_movements()

        super().__init__(
            analysis,
            interval,
            cases,
            max_ticks,
            log_scale,
            display,
            narrow_view,
            wide_view,
            chart_kwargs,
            **kwargs,
        )
        self._rulers: list = []

    @property
    def ChartCls(self) -> type[charts.TrendsChart]:
        return charts.TrendsChart

    @property
    def _chart_title(self) -> str:
        return self._analysis.symbol + " Trend Analysis"

    @property
    def _icon_row_top_handlers(self) -> list[Callable]:
        return [self._max_x_ticks, self._resize_chart, self.close]

    @property
    def current_case(self) -> Movement | MovementAlt | None:
        """Current selected case"""
        case = super().current_case
        if case is None:
            return None
        if typing.TYPE_CHECKING:
            assert isinstance(case, (Movement, MovementAlt))
        return case

    def _close_rulers(self):
        for ruler in self._rulers:
            ruler.close()
        self._rulers = []

    def _add_rulers(self):
        self._close_rulers()  # close any existing
        ohlc_mark = next(m for m in self.chart.figure.marks if isinstance(m, bq.OHLC))
        case = self.current_case
        assert case is not None
        self._rulers.append(
            HFixedRule(
                level=case.start_px,
                scales=self.chart.scales,
                figure=self.chart.figure,
                start=case.start.asm8,
                length=case.params["prd"],
                color="yellow",
                draggable=True,
                ordinal_values=list(ohlc_mark.x),
                stroke_width=5,
            )
        )

    def _ruler_handler(self, but: vu.IconBut, event: str, data: dict):
        if but.is_dark:
            return
        f = self._add_rulers if but.is_light else self._close_rulers
        f()
        # chain handlers...
        self.cases_controls_container.but_ruler_handler(but, event, data)

    def _create_cases_controls_container(self) -> v.Layout:
        controls = TrendControls()
        controls.but_show_all.on_event("click", self._show_all_but_handler)
        controls.but_next.on_event("click", self._select_next_case_handler)
        controls.but_prev.on_event("click", self._select_prev_case_handler)
        controls.but_narrow.on_event("click", self._narrow_view_handler)
        controls.but_wide.on_event("click", self._wide_view_handler)
        controls.but_ruler.on_event("click", self._ruler_handler)
        return controls


class TrendsGui(TrendsGuiBase):
    """GUI to visualise movements evaluated by `analy.Trends` class.

    Parameters
    ----------
    `prd`, `ext_break`, `ext_limit` and `min_bars` all as `analy.Trends`
    class.

    All other parameters as base class `TrendsGuiBase`.
    """

    def __init__(
        self,
        analysis: ma_analysis.Analysis,
        interval: mp.intervals.RowInterval,
        prd: int,
        ext_break: float,
        ext_limit: float,
        min_bars: int = 0,
        max_ticks: int | None = None,
        log_scale: bool = True,
        display: bool = True,
        **kwargs,
    ):
        trend_kwargs = {
            "prd": prd,
            "ext_break": ext_break,
            "ext_limit": ext_limit,
            "min_bars": min_bars,
        }
        super().__init__(
            analysis=analysis,
            interval=interval,
            trend_cls=Trends,
            trend_kwargs=trend_kwargs,
            max_ticks=max_ticks,
            log_scale=log_scale,
            display=display,
            narrow_view=prd,
            wide_view=prd * 3,
            **kwargs,
        )
        self.cases: Movements
        self.trends: Trends

    def _add_rulers(self):
        self._close_rulers()  # close any existing
        case = self.current_case
        assert case is not None
        is_adv = case.is_adv

        ohlc_mark = next(m for m in self.chart.figure.marks if isinstance(m, bq.OHLC))

        fctrs = self.trends.fctrs_pos_break if is_adv else self.trends.fctrs_neg_break
        self._rulers.append(
            market_analy.utils.bq_utils.TrendRule(
                x=case.start.asm8,
                y=case.start_px,
                length=case.params["prd"],
                factors=fctrs,
                scales=self.chart.scales,
                ordinal_values=list(ohlc_mark.x),
                figure=self.chart.figure,
                color="orange",
                draggable=True,
                stroke_width=3,
            )
        )

        fctrs = self.trends.fctrs_pos_limit if is_adv else self.trends.fctrs_neg_limit
        idx = -case.params["prd"]
        self._rulers.append(
            market_analy.utils.bq_utils.TrendRule(
                x=case.line_limit.index[idx].asm8,
                y=case.line_limit.iloc[idx],
                length=case.params["prd"],
                factors=fctrs,
                scales=self.chart.scales,
                ordinal_values=list(ohlc_mark.x),
                figure=self.chart.figure,
                color="skyblue",
                draggable=True,
                stroke_width=3,
            )
        )


class TrendsAltGui(TrendsGuiBase):
    """GUI to visualise movements evaluated by `analy_alt.Trends` class.

    Parameters
    ----------
    `prd`, `ext`, `rvr`, `grad`, `rvr_init` and `min_bars` all as
    `analy_alt.Trends` class.

    All other parameters as base class `TrendsGuiBase`.
    """

    @property
    def ChartCls(self) -> type[charts.TrendsAltChart]:
        return charts.TrendsAltChart

    def __init__(
        self,
        analysis: ma_analysis.Analysis,
        interval: mp.intervals.RowInterval,
        prd: int,
        ext: float,
        rvr: float | abc.Sequence[float] | RvrAlt,
        grad: float | None,
        rvr_init: float | tuple[float, float] | list[float],
        min_bars: int,
        max_ticks: int | None = None,
        log_scale: bool = True,
        display: bool = True,
        **kwargs,
    ):
        trend_kwargs = {
            "prd": prd,
            "ext": ext,
            "grad": grad,
            "rvr": rvr,
            "rvr_init": rvr_init,
            "min_bars": min_bars,
        }
        super().__init__(
            analysis=analysis,
            interval=interval,
            trend_cls=TrendsAlt,
            trend_kwargs=trend_kwargs,
            max_ticks=max_ticks,
            log_scale=log_scale,
            display=display,
            narrow_view=prd,
            wide_view=prd * 3,
            **kwargs,
        )
        self.cases: MovementsAlt
        self.trends: TrendsAlt



"""Trend Analysis.

Classes to evaluate and store trends. Trends evalutaed from OHLC data.

    Trends
        Broadly, a trend is considered to exist whenever a movement
        (advance or decline) is continually extended, by way of exceeding
        a limit line, whilst at the same time NOT exceeding a break line
        that underpins the trend. See `Trends.__doc__` for more specific
        documentation.

    TrendsAlt
        Broadly, a trend is considered to exist when the price has extended
        the high/low of a given period by a given percentage and has not
        subsequently 'lost' more than a given reversal limit. See
        `TrendsAlt.__doc__` for more specific documentation.
"""

from __future__ import annotations

import functools
import operator
import typing
from collections import abc
from dataclasses import dataclass
from functools import cached_property

import market_prices as mp
import numpy as np
import pandas as pd

from .movements import Movement, MovementAlt, Movements, MovementsAlt


class TrendParams(typing.TypedDict):
    """Trend Parameters.

    Trend parameters for a movement identified with `Trends`.
    """

    interval: mp.intervals.RowInterval
    prd: int
    ext_break: float
    ext_limit: float
    min_bars: int


class Trends:
    """Evaluate trends from OHLC data.

    Broadly, a trend is considered to exist whenever a movement (advance or
    decline) is continually extended, by way of exceeding a 'limit line'
    whilst at the same time NOT exceeding a 'break line' that underpins the
    trend. More specifically:

        A Movement is confirmed as having started whenever:
            Break Line over a given period (`prd`) has NOT been broken.
            For an advance/decline, the Break Line is defined as a
            rising/falling line with a given gradient (`prd` / `ext_break`)
            and which passes through the low/high of the bar on which the
            movement is considered to have started.

            Limit Line has been extended during a given period (`prd`).
            For an advance/decline, the Limit Line is defined as a
            rising/falling line with a given gradient (`prd` / `ext_limit`)
            and which passes through the low/high of the bar on which the
            advance/decline movement is considered to have started.

        Movement confirmed as having ended by way of:
            break, whenever the Break Line is exceeded. In this case the
            Break Line is defined as passing through the low/high of the
            most recent bar when the advance/decline fulfilled the
            conditions to be considered as having started (this may be the
            first time the movement was confirmed or a subsequent
            occasion).

            limit, whenever the Limit Line is not exceeded over the prior
            `prd`, where that Limit Line is defined as passing through the
            most recent high/low of the advance/decline.

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame with columns that include "high", "low" and "close". Any
        other columns will be ignored.

        DataFrame must be indexed with a `pd.DatetimeIndex`.

    interval : pd.Timedelta | str
        Interval that each row of `data` represents. Examples:

            pd.Timedelta(15, "min") (or "15min"): each row of data
            represents 15 minutes

            pd.Timedelta(1, "D") (or "1D"): for daily data.

    prd : int
        Period, in number of bars, over which the Break Line should not
        have been broken in order to declare a movement as having started.

        Also, the period over which the Limit Line must be extended to
        consider a movement as having started or to be continuing.

        See class documentation for definition of Break Line and
        Limit Line.

        Example, if each row of `data` (i.e. each bar) represents one
        session (i.e. if `interval` is "1D") then passing `prd` as 10 will
        set the period as 10 sessions, whilst if each bar represents an
        interval of 30 minutes (i.e. `interval` is "30min") then passing 10
        will set the period as 5 trading hours.

    ext_break
        The percentage change in the Break Line over `prd`. Accordingly,
        defines the gradient of the Break Line.

        Pass as a float, for example 0.005 for 0.5%.

        See class documentation for definition of 'Break Line'.

    ext_limit
        The percentage change in the Limit Line over `prd`. Accordingly,
        defines the gradient of the Limit Line.

        Pass as a float, for example 0.0025 for 0.25%.

        See class documentation for definition of 'Limit Line'.

    min_bars : int, default: 0
        The minimum number of bars that any movement should be comprised
        of.

    Notes
    -----
    NOTE: testing of `Trends` class is LIMITED.
    Tests are limited to verifying return as expected for certain
    combinations of parameters. In turn, 'what's expected' is as accepted
    from detailed visual inspection of the resulting trends. Class lacks,
    and requires, a test suite that verifies all possible returns of each
    of it's constituent methods, i.e. with parameters defined either side
    of boudaries to verify differences in return. Those tests should use
    dummy price data designed to the the boundaries.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        interval: mp.intervals.RowInterval,
        prd: int,
        ext_break: float,
        ext_limit: float,
        min_bars: int = 0,
    ):
        self.data = data.copy()
        self.interval = mp.intervals.to_ptinterval(interval)
        self.prd = prd

        self.ext_break = ext_break
        self.ext_limit = ext_limit
        self.min_bars = min_bars

        self.grad_break = ext_break / (prd - 1)
        self.grad_limit = ext_limit / (prd - 1)

        self.fctrs_pos_break = np.linspace(1, 1 + ext_break, prd)
        self.fctrs_neg_break = np.linspace(1, 1 - ext_break, prd)
        self.fctrs_pos_limit = np.linspace(1, 1 + ext_limit, prd)
        self.fctrs_neg_limit = np.linspace(1, 1 - ext_limit, prd)

    @property
    def params(self) -> TrendParams:
        """Parameters received by constructor."""
        return {
            "interval": self.interval,
            "prd": self.prd,
            "ext_break": self.ext_break,
            "ext_limit": self.ext_limit,
            "min_bars": self.min_bars,
        }

    @functools.lru_cache
    def base_intact(self, is_adv: bool) -> pd.Series:
        """Query if base was broken over any `self.prd` bars.

        Returns
        -------
        pd.Series
            index: as `self.data.index`
            dtype: bool

            Bool indicates if the base line was broken during the
            `self.prd` bars prior to the label. NB first `self.prd` bars
            are False.
        """

        def func(arr: np.ndarray) -> bool:
            """Query if base broken over an array shorter than `self.prd`."""
            fctrs = self.fctrs_pos_break if is_adv else self.fctrs_neg_break
            line = arr.iloc[0] * fctrs[: len(arr)]
            bv = (arr >= line) if is_adv else (arr <= line)
            return bv.all()

        col = self.data.low if is_adv else self.data.high
        bv = col.rolling(self.prd, min_periods=self.prd).apply(func)
        bv[bv.isna()] = 0
        return bv.astype("bool")

    @functools.lru_cache
    def limit_extended(self, is_adv: bool) -> pd.Series:
        """Query if limit is extended over any `self.prd` bars.

        Returns
        -------
        pd.Series
            index: as `self.data.index`
            dtype: bool

            Bool indicates if the limit line was extended during the
            `self.prd` bars prior to the label. NB first `self.prd` bars
            are False.
        """

        def func(arr: np.ndarray) -> bool:
            """Query if limit extended over array shorter than `self.prd`."""
            fctrs = self.fctrs_pos_limit if is_adv else self.fctrs_neg_limit
            line = arr.iloc[0] * fctrs[: len(arr)]
            bv = (arr > line) if is_adv else (arr < line)
            return bv.any()

        col = self.data.high if is_adv else self.data.low
        bv = col.rolling(self.prd, min_periods=self.prd).apply(func)
        bv[bv.isna()] = 0
        return bv.astype("bool")

    @functools.lru_cache
    def start_confs_all(self, is_adv: bool) -> pd.DatetimeIndex:
        """All bars on which a movement could be confirmed as having started."""
        bv = self.base_intact(is_adv) & self.limit_extended(is_adv)
        return bv[bv].index

    def start_from_start_conf(self, start_conf: pd.Timestamp) -> pd.Timestamp:
        """Get start bar from start confirmed bar."""
        idx_start_conf = self.data.index.get_loc(start_conf)
        idx_start = idx_start_conf - self.prd + 1
        return self.data.index[idx_start]

    def start_conf_from_start(self, start: pd.Timestamp) -> pd.Timestamp | None:
        """Get start confrirmed bar corresponding with a start bar.

        Returns None if `conf_start` would be beyond the end of available
        data.
        """
        idx_start = self.data.index.get_loc(start)
        idx_start_conf = idx_start + self.prd - 1
        if idx_start_conf > len(self.data):
            return None
        return self.data.index[idx_start_conf]

    def get_line(self, subset: pd.Series, is_adv: bool, limit=True) -> pd.Series:
        """Get a base or limit line for a subset of the data.

        Parameters
        ----------
        subset
            Continuous (over rows) subset of `self.data`, either 'high' or
            'low' column.

        is_adv
            True: if movement is an advance.
            False: if movement is a decline.

        limit
            True: return a limit line.
            False: retrun a break line.

        Returns
        -------
        line : pd.Series
            Series with index as subset and values delineating a line with
            gradient `self.grad_break` or `self.grad_limit` passing through
            `subset`[0].
        """
        op = operator.add if is_adv else operator.sub
        grad = self.grad_limit if limit else self.grad_break
        fctrs = np.array([op(1, (i * grad)) for i in range(len(subset))])
        return pd.Series(fctrs * subset.iloc[0], index=subset.index)

    def get_limit_line(
        self, start: pd.Timestamp, end: pd.Timestamp, is_adv: bool
    ) -> pd.Series:
        col = self.data.high if is_adv else self.data.low
        subset = col[start:end]

        srss = []
        while True:
            line = self.get_line(subset, is_adv)
            bv = subset > line if is_adv else subset < line
            idx = bv.argmax() if bv.any() else None
            srss.append(line[:idx])
            if idx is None or idx + 1 == len(bv):
                break
            subset = subset[idx:]

        return pd.concat(srss)

    def get_limit(
        self, start_conf: pd.Timestamp, is_adv: bool
    ) -> tuple[pd.Timestamp | None, float | None, pd.Series]:
        """Get bar and price when price has failed to extend limit.

        Parameters
        ----------
        start_conf
            Bar when movement was confirmed.

        is_adv
            True: Movement represents an advance.
            False: Movemment represents a decline.

        Returns
        -------
        bar : pd.Timestamp | None
            First bar, following `start_conf`, when the price failed to
            extend the limit line over the prior `self.prd`.

            None if price did not fail to extend the limit line as at the
            end of the available data.

        break_px : float | None
            Close price of `bar` (as bar is included within `self.prd`, all
            of day should be included as it had this period available to
            try and extend the limit).

            None if price did not fail to extend the limit line as at the
            end of the available data.

        limit_line : pd.Series
            Limit line from movement start through `bar`.
        """
        bv = self.limit_extended(is_adv).loc[start_conf:]
        if bv.all():
            bar, px = None, None
            bar_ = self.data.index[-1]
        else:
            bar = bar_ = bv.index[bv.argmin()]
            px = self.data.loc[bar].close

        start = self.start_from_start_conf(start_conf)
        line = self.get_limit_line(start, bar_, is_adv)
        return bar, px, line

    def get_break(
        self, start_conf: pd.Timestamp, is_adv: bool
    ) -> tuple[pd.Timestamp | None, float | None, pd.Series]:
        """Get bar and price when price first breaks break line.

        Gets bar when price breaks base line for the first time
        following a movement's confirmation.

        Parameters
        ----------
        start_conf
            Bar on which start of movement confirmed.

        is_adv
            True: movement for which require break is an advance.
            False: movement for which require break is a decline.

        Returns
        -------
        break_ : pd.Timestamp | None
            Bar that movement, with confirmed start as `start_conf`,
            subsequently broke the break line for the first time.

            None if price did not break the break line by the end of the
            available data.

        break_px : float | None
            Price when break line broken. This will be the value of the
            break line or the bar open if the open exceeded the break line.

            None if price did not exceed break line by the end of the
            available data.

        break_line : pd.Series
            Break line from movement start through `break_`.
        """
        start_confs_all = self.start_confs_all(is_adv)
        idx = start_confs_all.get_loc(start_conf)
        start_confs = start_confs_all[idx:]

        col = self.data.low if is_adv else self.data.high
        frms = pd.Series(start_confs).apply(self.start_from_start_conf)
        tos = frms.shift(-1)
        srss = []
        for frm, to in zip(frms, tos):
            if pd.isna(to):
                to = None
            subset = col[frm:to]
            line = self.get_line(subset, is_adv, limit=False)
            bv = subset < line if is_adv else subset > line
            if not bv.any():
                srss.append(line[1:])  # miss first so they don't overlap
                continue

            idx = bv.argmax()
            srss.append(line[1 : idx + 1])
            line_break = pd.concat(srss)
            break_ = line_break.index[-1]
            f = min if is_adv else max
            break_px = f(line_break.iloc[-1], self.data.loc[break_].open)
            return break_, break_px, line_break

        return None, None, pd.concat(srss)

    def get_end_conf(
        self, start_conf: pd.Timestamp, is_adv: bool
    ) -> tuple[pd.Timestamp | None, float | None, bool | None, pd.Series, pd.Series]:
        """Get bar and price when a movement is confirmed as having ended.

        Returns
        -------
        tuple(None, None, None, pd.Series, pd.Series) if movement end had
        not been confirmed as at end of available data, otherwise:

        end_conf : pd.Timestamp
            Bar on which movement end confirmed.

        end_conf_px : float
            Price when movement end was confirmed.

        by_break : bool
            True if movement ended as a result of breaking the base line
            on `end`.

            False if movement ended as a result of failing to extend the
            limit line over the `self.prd` bars through to `end`.

        line_break : pd.Series
            Break line from movement start through to bar when price broke
            the break line, or to last bar if break line was not broken as
            at the end of available data.

        line_limit : pd.Series
            Limit line from movement start through to first bar when price
            has failed to extend limit line over the prior `self.prd` bars,
            or through to last bar if as at the end of the available data
            the price had not failed to exceed the limit line.
        """
        end_break, end_px_break, line_break = self.get_break(start_conf, is_adv)
        end_limit, end_px_limit, line_limit = self.get_limit(start_conf, is_adv)
        if end_break is None and end_limit is None:
            return None, None, None, line_break, line_limit
        if end_break is None:
            return end_limit, end_px_limit, False, line_break, line_limit
        if end_limit is None:
            return end_break, end_px_break, True, line_break, line_limit
        # if same bar then break will occur first as limit happens at bar close
        by_break = end_break <= end_limit
        px = end_px_break if by_break else end_px_limit
        return min(end_break, end_limit), px, by_break, line_break, line_limit

    def get_end(
        self, start: pd.Timestamp, end_conf: pd.Timestamp | None, is_adv: bool
    ) -> tuple[pd.Timestamp, float]:
        """Get movement end bar and price."""
        col = self.data.high if is_adv else self.data.low
        subset = col.loc[start:end_conf]
        _end_idx = subset.argmax() if is_adv else subset.argmin()
        end = subset.index[_end_idx]
        end_px = subset.max() if is_adv else subset.min()
        return end, end_px

    def get_start(
        self, start_conf: pd.Timestamp, is_adv: bool
    ) -> tuple[pd.Timestamp, float]:
        """Get movement start bar and price."""
        start = self.start_from_start_conf(start_conf)
        row = self.data.loc[start]
        return start, row.low if is_adv else row.high

    def get_duration(self, start: pd.Timestamp, end: pd.Timestamp) -> int:
        idx_start = self.data.index.get_loc(start)
        idx_end = self.data.index.get_loc(end)
        return idx_end - idx_start + 1

    def get_movement(self, start_conf: pd.Timestamp, is_adv: bool) -> Movement:
        """Get movement corresponding with a given confirmed start bar."""
        end_conf, end_conf_px, by_break, line_break, line_limit = self.get_end_conf(
            start_conf, is_adv
        )
        start, start_px = self.get_start(start_conf, is_adv)
        end, end_px = self.get_end(start, end_conf, is_adv)
        start_conf_px = self.data.loc[start_conf].close

        return Movement(
            is_adv,
            start,
            start_px,
            start_conf,
            start_conf_px,
            end,
            end_px,
            end_conf,
            end_conf_px,
            self.get_duration(start, end),
            self.params,
            line_break,
            line_limit,
            by_break,
        )

    def _get_movements(self, is_adv: bool) -> list[Movement]:
        """Return list of movements of a given direction."""
        moves = []
        start_confs_all = self.start_confs_all(is_adv)
        while True:
            move = self.get_movement(start_confs_all[0], is_adv)
            if move.duration < self.min_bars:
                start_confs_all = start_confs_all[1:]
                if start_confs_all.empty:
                    break
                continue
            moves.append(move)
            if move.end is None:
                break
            minconfstart = self.start_conf_from_start(move.end)
            if minconfstart is None:
                break
            idx = start_confs_all.get_slice_bound(minconfstart, side="right")
            start_confs_all = start_confs_all[idx:]
            if start_confs_all.empty:
                break
        return moves

    def get_movements(self) -> Movements:
        """Return all movements."""
        advs = self._get_movements(True)
        decs = self._get_movements(False)
        moves = sorted(advs + decs)

        # sanity check that moves do not overlap except end/start
        prev_move: Movement | None = None
        for move in moves:
            if prev_move is not None:
                assert prev_move.end <= move.start
            prev_move = move

        return Movements(tuple(moves), self.data, self.interval)


@dataclass(frozen=True)
class RvrAlt:
    """Parameters for ALTERNATIVE definition of reversal percent limits.

    Parameters provide for defining reversal percent limits according to
    the maximum extent of the movement as a factor of the instrument's
    'usual' change over `prd` number of bars. The 'usual' change is
    evaluated:
        For advances, as the median positive "change over prd" over
        the full extent of the data set.

        For declines, as the median negative "change over prd" over
        the full extent of the data set.

    Parameters
    ----------
    multiples : list[int | float]
        Multiples of the 'usual' change up to which to the reversal percent
        limit defined by the corresponding index of `rvr` should be
        applied (see `rvr`).

    rvr : list[float]
        Reversal percent limits that should be applied when the movement's
        maximum extent, as a multiple of the 'usual' change, is less than
        the value of the corresponding index of `multiples`. List should
        have length one greater than the length of `multiples` with the last
        value defining the reversal limit to apply beyond the last multiple
        of `multiples`.

        For example, if the 'usual' positive change over prd is 2% and...
            multiples = (1.5, 2.5, 4, 6)
            rvr = (0.8, 0.6, 0.4, 0.2, 0.15)
        ...then the reversal limit would be 0.8* until an advance exceeds 3%
        (i.e. 2% * the corresponding 1.5 factor), thereafter 0.6 until an
        advance exceeds 5% (2% * 2.5), thereafter 0.4 until the advance
        exceeds 8%, thereafter 0.2 until the advance exceeds 12% and
        thereafter 0.15.

        * i.e. the advance would be considered over if the movement were to
        'give back' 80% of it's maximum extent since the movement was
        confirmed.

    start_limit : tuple[float, int] | None, default: None
        Defines a minimum reversal limit to apply over the start of the
        movement. 2-tuple describing:
            [0] : float. Minimum reversal limit that should be applied
            over the period described by [1].

            [1] : int. Number of bars, from and inclusive of the bar on
            which the movement was confirmed, over which the start limit
            should apply.

    end_limit : tuple[int, float] | None, default: None
        Defines a maximum reversal limit to apply over the 'end' of the
        movement. 2-tuple describing:
            [0] : float. Maximum reversal limit that should be applied
            over the period described by [1].

            [1] : int. Number of bars, from and inclusive of the bar on
            which the movement was confirmed, after which the end limit
            should be applied. If `start_limit` passed then `end_limit`[1]
            must by higher than `start_limit`[1].
    """

    multiples: list[int | float]
    rvr: list[float]
    start_limit: tuple[float, int] | None = None
    end_limit: tuple[float, int] | None = None

    def __post_init__(self):
        if len(self.rvr) != len(self.multiples) + 1:
            raise ValueError(
                "`rvr` must have a length one greater than the length of `multiples`."
            )
        if self.start_limit is not None and self.end_limit is not None:
            if not self.end_limit[1] > self.start_limit[1]:
                raise ValueError("`end_limit` must commence after `start_limit`.")


class TrendAltParams(typing.TypedDict):
    """Trend Parameters.

    Trend parameters for a movement identified with `Trends`.
    """

    interval: mp.intervals.RowInterval
    prd: int
    ext: float
    rvr: list | RvrAlt
    rvr_init: np.ndarray
    grad: float | None
    min_bars: int


class TrendsAlt:
    """Evaluate trends over OHLC data.

    --Movement Start Confirmed Bar--
    A movement is confirmed as having started on a bar if price extends
    high/low of prior `prd` by `ext`. Specifically, an advance/decline is
    confirmed if the price exceeds, for the first time, a Limit Line which
    increases/decreases linearly and is defined:
        TO the bar on which movement could be confirmed, at the price
        that the movement could be confirmed.

        FROM `prd` bars prior to the 'TO' bar, at a price of `ext` %
        below/above the 'TO' price.

    --Start Bar--
    The bar on which an advance/decline starts is, in the first instance,
    considered as the low/high of the `prd` bars prior to the bar on which
    the movement was confirmed. If, in accordance with the initial reversal
    limits (`rvr_init`), the movement would be considered to have ended
    before it was confirmed then the start bar will be advanced to the
    first subsequent low/high from where the movement would not be
    considered to have ended prior to the start confirmation bar.

    Separately, see 'Notes' section for notes for circumstances that can
    result in a start bar being defined later than would seen most
    appropriate.

    --Movement End Confirmed Bar--
    A movement is confirmed as having ended by way of:
        Consolidation - on any bar when the criteria to consider a movement
        to have started is not met over any of the prior `prd` bars.

        Reversal - in accordance with the reversal limits set by `rvr`.

    --End Bar--
    The bar on which an advance/decline ends is, in the first instance,
    considered as the high/low of the `prd` bars prior to the bar on which
    the movement end was confirmed. If the movement ended by way of
    consolidation then this end bar will be moved back to the earliest of
    any bar that exceeds the End Establishment Line (EEL) over the prior
    `prd` bars (see `grad` parameter).

    Parameters
    ----------
    data : pd.DataFrame
        DataFrame with columns that include "high", "low" and "close". Any
        other columns will be ignored.

        DataFrame must be indexed with a `pd.DatetimeIndex`.

    interval : pd.Timedelta | str
        Interval that each row of `data` represents. Examples:

            pd.Timedelta(15, "min") (or "15min"): each row of data
            represents 15 minutes

            pd.Timedelta(1, "D") (or "1D"): for daily data.

    prd : int
        Period over which extreme should have been advanced by no less than
        `ext` in order for a movement to be considered to having started
        or be continuing. Passed in terms of number of rows of `data` (i.e.
        bars on a OHLC chart). For example, if each row of `data`
        represents one session (i.e. if `interval` is "1D") then passing 10
        will set the period as 10 sessions, whilst if each row represents
        an interval of 30 minutes (i.e. `interval` is "30min") then passing
        10 will set the period as 5 hours.

    ext : float
        If advance/decline fails to extend the high/low of the previous
        `prd` by `ext` then movement will be considered to have ended
        by way of consolidation. Pass as a float, for example 0.005 for
        0.5%. Extensions of the high/low within `prd` are considered on
        a pro-rata basis (see 'Limit Line' comments within the introduction
        to the class doc).

    rvr : float | abc.Sequence[float] | RvrAlt
        When passing as a float or sequence of float:

            A movement will be considered to have ended by way of reversal
            on any counter movement greater than `rvr` percent of the trend
            to date.

            Pass as a float to define as a single static percentage. For
            example, 0.5 will result in a movement that's advanced 12% to
            be continuing if it were to fall 5.9% from it's high, although
            the movement would be considered to have ended by way of
            reversal if the price were to fall to 6% under the movement
            high (i.e. 50% of the movement would have been given up).

            Alternatively, pass as a sequence of float to define a
            differing percentage for every bar following the bar when the
            movement is confirmed, for example [0.8, 0.8, 0.8, 0.7, 0.7,
            0.6. 0.6, 0.5, 0.5, 0.5, 0.5, 0.5, 0.4, 0.4, 0.4, 0.4 0.4, 0.3]
            would consider the advance to have ended by way of reversal if
            during the three bars following the movements confirmation the
            price 'gave up' more than 80% of the trend to date, whilst over
            the following 2 bars only 70% of the trend would need given up
            in order for the trend to be confirmed as having given way to
            reversal, 60% over the next 2 bars, 50% over the next 5, 40%
            over the next 5 and then 30% from then on (the last value will
            be considered static henceforth).

        If passing as RvrAlt:

            A movement will be considered to have ended by way of reversal
            on any counter movement greater than reversal limits evaluated
            according to the extent of the movement. The extent of the
            movement is evaluated in terms of a "multiple of the
            instrument's usual change over `prd`", where 'usual' is
            evaluated as:
                For advances, the median positive "change over `prd`" over
                the full extent of the data set.

                For declines, the median negative "change over `prd`" over
                the full extent of the data set.

            The `RvrAlt` dataclass provides for defining the multiples and
            corresponding reversal limits together with, optionally,
            minimum / maximum reversal limits to apply over the start / end
            of a movement.

    rvr_init : float | tuple | list
        Reversal limits for the period prior to confirmation that a
        movement has started. These limits determine the start bar such
        that from the bar after the start through to the confirmed start
        the price will not have reversed by more than `rvr_init`.

        Similar to `rvr`, pass as a float to define a static percentage to
        apply throughout this initial period, or pass as a list to define
        a differing percentage for every bar following the start bar. If
        the list is longer than the number of bars from the bar after the
        assessed start through to the confirmed start then the excess
        elements at the end of the list will be ignored. If the list is
        shorter than the number of bars then the last element of the list
        will be repeated.

        Alternatively, pass as a 2-tuple of floats to define a value for
        the bar after the start and a value for a bar `prd` - 1 bars
        later. A list of length `prd` will be created with the values in
        between evaluated by linear interpolation. Whenever this list is
        longer than the number of bars between an assessed start and the
        confirmed start the excess elements at the end of the list will be
        ignored.

    grad : float | None
        Defines gradient of End Establishment Line (EEL).

        Default (if None), `ext` / (`prd` * 2)

        A movement's end bar is determined as the earliest bar that the
        price touches the EEL, where the EEL is defined as a line that:

            - end on the high/low of the advance/decline as registered
            over the `prd` bars prior to the bar on which the movement
            end is confirmed.

            - has a length `prd` + 1 bars.

            - has a gradient such that for an advance/decline the line
            rises/falls by `grad` every bar.

    min_bars : int, default: 0
        The minimum number of bars that should comprise any movement.

    Notes
    -----
    The start should be earlier!
    Sometimes the start bar of a trend appears later than would seem most
    appropriate. For example, the start of an advance might be set a couple
    of bars after a lower low. This is usually because subsequently the
    movement is not extended by `ext` over `prd` bars - the criteria to
    confirm a movement's end cannot be fulfiled 'during' an ongoing
    movement! Integrity is important.

    If a notable number of movements are 'starting late', consider
    investigating if there are more suitable parameters to represent the
    instrument.

    NOTE: testing of `Trends` class is LIMITED.
    Tests are limited to verifying return as expected for certain
    combinations of parameters. In turn, 'what's expected' is as accepted
    from detailed visual inspection of the resulting trends. Class lacks,
    and requires, a test suite that verifies all possible returns of each
    of it's constituent methods, i.e. with parameters defined either side
    of boudaries to verify differences in return. Those tests should use
    dummy price data designed to the the boundaries.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        interval: mp.intervals.RowInterval,
        prd: int,
        ext: float,
        rvr: float | abc.Sequence[float] | RvrAlt,
        rvr_init: float | tuple[float, float] | list[float],
        grad: float | None,
        min_bars: int = 0,
    ):
        self.data = data.copy()
        self.interval = mp.intervals.to_ptinterval(interval)
        self.prd = prd
        self.ext = ext
        self.grad = grad if grad is not None else self.ext / (self.prd * 2)
        self.min_bars = min_bars

        self.rvr: list[float] | RvrAlt
        if isinstance(rvr, RvrAlt):
            self.rvr = rvr
        else:
            self.rvr = [rvr] if isinstance(rvr, (float, int)) else list(rvr)

        self.rvr_init: np.ndarray  # set as array of len no less than prd
        if isinstance(rvr_init, tuple):
            self.rvr_init = np.linspace(rvr_init[0], rvr_init[-1], prd)
        elif isinstance(rvr_init, (float, int)):
            self.rvr_init = np.array([rvr_init] * prd)
        else:
            if len(rvr_init) < prd:
                rvr_init = rvr_init + ([rvr_init[-1]] * (prd - len(rvr_init)))
            self.rvr_init = np.array(rvr_init)

    @property
    def params(self) -> TrendAltParams:
        """Parameters received by constructor."""
        return {
            "interval": self.interval,
            "prd": self.prd,
            "ext": self.ext,
            "rvr": self.rvr,
            "rvr_init": self.rvr_init,
            "grad": self.grad,
            "min_bars": self.min_bars,
        }

    @cached_property
    def factors_break_pos(self) -> np.ndarray:
        """Positive factors to identify break, increasing left to right."""
        return np.concatenate(
            [np.array([1]), np.linspace(1, 1 + self.ext, self.prd - 1)]
        )

    @cached_property
    def factors_break_neg(self) -> np.ndarray:
        """Negative factors to identify break, decreasing left to right."""
        return np.concatenate(
            [np.array([1]), np.linspace(1, 1 - self.ext, self.prd - 1)]
        )

    @cached_property
    def factors_break_pos_rev(self) -> np.ndarray:
        """Positive factors to identify break, increasing right to left."""
        return self.factors_break_pos[::-1]

    @cached_property
    def factors_break_neg_rev(self) -> np.ndarray:
        """Negative factors to identify break, decreasing right to left."""
        return self.factors_break_neg[::-1]

    @cached_property
    def grad_factors_pos(self) -> list[float]:
        """Positive gradient factors, increasing from left to right."""
        return [1, 1] + [1 + (x * self.grad) for x in range(3, self.prd + 1)]

    @cached_property
    def grad_factors_neg(self) -> list[float]:
        """Negative gradient factors, decreasing from left to right."""
        return [1, 1] + [1 - (x * self.grad) for x in range(3, self.prd + 1)]

    @cached_property
    def grad_factors_pos_rev(self) -> list[float]:
        """Positive gradient factors, increasing right to left."""
        return list(reversed(self.grad_factors_pos))

    @cached_property
    def grad_factors_neg_rev(self) -> list[float]:
        """Negative gradient factors, increasing right to left."""
        return list(reversed(self.grad_factors_neg))

    @staticmethod
    def get_reversals(
        px_start: float,
        df: pd.DataFrame,
        is_adv: bool,
        limits: np.ndarray | None = None,
    ) -> pd.Series | tuple[pd.Series, pd.Series]:
        """Get percentage reversals from high/low relative to a start price.

        Parameters
        ----------
        px_start
            Start price against which to measure movements and subsequent
            reversals.

        df
            pd.DataFrame covering period to assess. First row should be the
            row that immediately follows the row corresponding with
            `px_start`.

        is_adv
            True: movement represents an advance
            False: movement represents a decline

        limits
            List of same length as `df` describing reversal limits for each
            bar.

            Optional, if passed then return will include a pd.Series
            describing a line that tracks these limits.

        Returns
        -------
        rvr : pd.Series
            pd.Series with index as `df` and values as reversal from
            current high/low.

        line : pd.Series
            Only returned if `limits` passed.

            pd.Series, with index as `df` describing a line that tracks
            `limits`.
        """
        end_cur = df.high.cummax() if is_adv else df.low.cummin()
        end_rvr = df.low.copy() if is_adv else df.high.copy()
        # for intervals where end_cur has been moved out, not possible to
        # know if the high/low came before the new low/high or afterwards,
        # given which assumes counter movement reached only interval close.
        bv = end_cur != end_cur.shift(1)
        end_rvr[bv] = df.close[bv]
        chg = end_cur - px_start if is_adv else px_start - end_cur
        chg_rvr = (end_rvr - end_cur).abs()
        rvr = chg_rvr / chg
        if limits is None:
            return rvr
        line = px_start + (chg * ((1 - limits) * (1 if is_adv else -1)))
        return rvr, line

    def get_start(
        self,
        minstart: pd.Timestamp,
        start_conf: pd.Timestamp,
        is_adv: bool,
    ) -> tuple[pd.Timestamp, float, pd.Series]:
        """Get the start date of a potential movement.

        Parameters
        ----------
        minstart
            Returned start will be no earlier than `minstart`.

        start_conf
            The bar on which the start of the movement is confirmed.

        is_adv
            True: movement would represent an advance.
            False: movement would represent a decline.

        Returns
        -------
        3-tuple of [pd.Timestamp | pd.Interval, float | pd.Series]
            [0] Bar that would represent start of movement.
            [1] Price at what would be start of movement.
            [2] Start establishement line.

        Notes
        -----
        Initially assumes start as the lowest/highest price of the
        `self.prd` period prior to the bar on which the movement was
        confirmed.

        If with this start the movement would fulfil the critiera to
        confirm the end of the movememnt before the confirmed start then
        moves the start forwards to a bar following which the criteria to
        confrim the end of the movement is not met.
        """
        index = self.data.index
        idx_conf = index.get_loc(start_conf)
        idx_min = 0 if minstart not in index else index.get_loc(minstart)

        idx_frm = max(idx_min, idx_conf - self.prd)
        subset = self.data.iloc[idx_frm : idx_conf + 1]
        _idx_start = subset.low.argmin() if is_adv else subset.high.argmax()
        row = subset.iloc[_idx_start]
        idx_start = index.get_loc(row.name)

        px_start = row.low if is_adv else row.high
        subset = self.data.iloc[idx_start + 1 : idx_conf + 1]  # <= prd len subset
        rvr_limits = self.rvr_init[: len(subset)]
        rvr, line = self.get_reversals(px_start, subset, is_adv, rvr_limits)
        bv = rvr.values > rvr_limits
        while bv.any():
            _idx_start = bv.argmax()
            row = subset.iloc[_idx_start]
            px_start = row.low if is_adv else row.high
            subset = subset[_idx_start + 1 :]
            rvr_limits = self.rvr_init[: len(subset)]
            rvr, line_ = self.get_reversals(px_start, subset, is_adv, rvr_limits)
            line = pd.concat([line[: -len(line_)], line_])
            bv = rvr.values > rvr_limits

        return row.name, row.low if is_adv else row.high, line

    def find_movement(
        self, df: pd.DataFrame, minstart: pd.Timestamp, minstart_conf: pd.Timestamp
    ) -> (
        tuple[bool, pd.Timestamp, float, pd.Series, pd.Timestamp, float, pd.Series]
        | None
    ):
        """Identify the first movement to occur.

        Identifies the first movement to be confirmed no earlier than
        `minstart_conf`.

        An advance/decline is considered to be confirmed on a bar when the
        price is greater/lesser than all values of the prior `self.prd` - 1
        bars and has extended those prior highs/lows by a greater degree
        than would be considered as consolidation (i.e. as determined by
        having exceeded the Limit Line - see introduction to class doc).

        Parameters
        ----------
        df
            pd.DataFrame with rows representing bar interval and columns as
            OHLC. Method makes no changes to this dataframe.

        minstart
            Returned movement will have a start bar no earlier than
            `minstart`.

        minstart_conf
            Returned movement will have a confirmation bar no earlier than
            `minstart_conf`. Cannot be earlier than `minstart`.

        Returns
        -------
        None | 7-tuple of
            [bool, pd.Timestamp, float, pd.Series, pd.Timestamp, float,
            pd.Series]:
            [0] Movement direction, True if an advance, False if a decline.
            [1] Bar representing movement's start.
            [2] Price at start of movement.
            [3] Start establishement line.
            [4] Bar when movement broke consolidation of prior `self.prd`.
            [5] Break price of prior consolidation.
            [6] Line indicating limit of broken consolidation range.

            None if no movement is identified.
        """
        new_high = df.high == df.high.rolling(self.prd, min_periods=0).max()
        new_low = df.low == df.low.rolling(self.prd, min_periods=0).min()

        # only consider as a movement if register new prd high/low but NOT if register
        # both a new prd high and new prd low...possible for both to register on the
        # same interval if the range of the last prd falls fully within the interval's
        # range. In this case wait for the trend to be clarified.
        new_high_xor_low = (new_high & ~new_low) | (new_low & ~new_high)
        new_high_xor_lows = df.index[
            new_high_xor_low & (new_high_xor_low.index >= minstart_conf)
        ]

        for high_xor_low in new_high_xor_lows:
            is_adv = new_high[high_xor_low]
            is_dec = not is_adv
            idx_conf = df.index.get_loc(high_xor_low)
            row = df.loc[high_xor_low]
            px = row.high if is_adv else row.low
            fctrs = self.factors_break_pos_rev if is_adv else self.factors_break_neg_rev
            subset = df.iloc[idx_conf - self.prd : idx_conf]
            vals = fctrs * (subset.high if is_adv else subset.low)

            if (is_adv and not (vals > px).any()) or (is_dec and not (vals < px).any()):
                start, start_px, sel = self.get_start(minstart, high_xor_low, is_adv)
                idx_start = df.index.get_loc(start)
                if idx_conf - idx_start + 1 < self.min_bars:
                    continue

                break_px = vals.max() if is_adv else vals.min()
                conf_px = max(break_px, row.open) if is_adv else min(break_px, row.open)
                f = self.factors_break_neg_rev if is_adv else self.factors_break_pos_rev
                line = pd.Series(f, subset.index) * break_px
                line[high_xor_low] = break_px
                return is_adv, start, start_px, sel, high_xor_low, conf_px, line
        return None

    def _find_adv(
        self, df: pd.DataFrame, minstart_conf: pd.Timestamp | None = None
    ) -> tuple[pd.Timestamp | pd.Interval, float, pd.Series] | tuple[None, None, None]:
        """Identify the first advance to occur.

        Returns
        -------
        3-tuple of
            [pd.Timestamp, float, pd.Series] | [None, None, None]:
            [0] Bar of `df` when movement broke consolidation of the prior
                `self.prd`.
            [1] break price of prior consolidation (or nearest traded
                price).
            [2] line indicating limit of broken consolidation range.

            (None, None, None) if no advance is identified.

        Notes
        -----
        Skimmed down version of `.find_movement` concerned only with
        identifying the next advance.
        """
        new_high = df.high == df.high.rolling(self.prd, min_periods=0).max()
        new_highs = df.index[new_high]
        if minstart_conf is not None:
            new_highs = new_highs[new_highs >= minstart_conf]

        for high in new_highs:
            row = df.loc[high]
            px = row.high
            idx_conf = df.index.get_loc(high)
            subset = df.iloc[idx_conf - self.prd : idx_conf]
            vals = self.factors_break_pos_rev * subset.high
            if not (vals > px).any():
                start, _, _ = self.get_start(minstart_conf, high, True)
                idx_start = df.index.get_loc(start)
                if idx_conf - idx_start + 1 < self.min_bars:
                    continue
                break_px = vals.max()
                fctrs = self.factors_break_neg_rev
                line = pd.Series(fctrs, subset.index) * break_px
                line[high] = break_px
                return high, max(break_px, row.open), line
        return None, None, None

    def _find_dec(
        self, df: pd.DataFrame, minstart_conf: pd.Timestamp | None = None
    ) -> tuple[pd.Timestamp, float, pd.Series] | tuple[None, None, None]:
        """Identify the first decline to occur.

        Returns
        -------
        3-tuple of
            [pd.Timestamp, float, pd.Series] | [None, None, None]:
            [0] Bar of `df` when movement broke consolidation of the prior
                `self.prd`.
            [1] break price of prior consolidation (or nearest traded
                price).
            [2] line indicating limit of broken consolidation range.

            (None, None, None) if no decline is identified.

        Notes
        -----
        Skimmed down version of `.find_movement` concerned only with
        identifying the next decline.
        """
        new_low = df.low == df.low.rolling(self.prd, min_periods=0).min()
        new_lows = df.index[new_low]
        if minstart_conf is not None:
            new_lows = new_lows[new_lows >= minstart_conf]

        for low in new_lows:
            row = df.loc[low]
            px = row.low
            idx_conf = df.index.get_loc(low)
            subset = df.iloc[idx_conf - self.prd : idx_conf]
            vals = self.factors_break_neg_rev * subset.low
            if not (vals < px).any():
                start, _, _ = self.get_start(minstart_conf, low, False)
                idx_start = df.index.get_loc(start)
                if idx_conf - idx_start + 1 < self.min_bars:
                    continue
                break_px = vals.min()
                fctrs = self.factors_break_pos_rev
                line = pd.Series(fctrs, subset.index) * break_px
                line[low] = break_px
                return low, min(break_px, row.open), line
        return None, None, None

    def _conf_end_rvr_opp(
        self, is_adv: bool, conf_start: pd.Timestamp
    ) -> tuple[pd.Timestamp, float, pd.Series] | tuple[None, None, None]:
        """Get info for next movement in opposing direction.

        Get info for when a movement would end as a result of giving way to
        a movement being confirmed in the opposite direction.

        Parameters
        ----------
        conf_start
            Interval when movement confirmed.

        is_adv
            True: movement represents an advance.
            false: movement represents a decline.

        Returns
        -------
        3-tuple of
            [pd.Timestamp, float, pd.Series] | [None, None, None]:
            [0] Bar when opposing movement confirmed.
            [1] Price when opposing movement confirmed.
            [2] Line indicating limit of broken consolidation range.

            (None, None, None) if no opposing movement is identified.
        """
        i = self.data.index.get_loc(conf_start) - self.prd
        df = self.data.iloc[i:]
        f = self._find_dec if is_adv else self._find_adv
        return f(df, conf_start)

    def _prd_chgs(self) -> pd.Series:
        df = self.data
        opens = df.open
        closes = df.close.shift(-self.prd + 1)
        return (closes - opens) / opens

    @cached_property
    def _av_chg_pos(self) -> float:
        srs = self._prd_chgs()
        return srs[srs > 0].median()

    @cached_property
    def _av_chg_neg(self) -> float:
        srs = self._prd_chgs()
        return srs[srs < 0].median()

    def _get_rvr_arr(
        self, is_adv: bool, end_cur: pd.Series, px_start: float
    ) -> np.ndarray:
        """Get array of reversal percent limits.

        Array will be as long as `end_cur` and correspond with the bars as
        described by the index of `end_cur`.

        Parameters
        ----------
        is_adv
            True: movement being evaluated represents an advance.
            false: movement being evaluated represents a decline.

        end_cur : pd.Series
            Index, dtype 'datetime64[ns]', representing bars over which
            require reversal limits. First row should represent the bar on
            which the movement was confirmed.

        px_start : float
            Price at start of movement.
        """
        if not isinstance(self.rvr, RvrAlt):
            arr = np.array([self.rvr[-1]] * len(end_cur))
            lngth = min(len(self.rvr), len(end_cur))
            arr[:lngth] = self.rvr[:lngth]
            return arr

        av = self._av_chg_pos if is_adv else abs(self._av_chg_neg)
        breaks = [0] + [m * av for m in self.rvr.multiples] + [np.inf]
        index = pd.IntervalIndex.from_breaks(breaks, closed="left")
        srs = pd.Series(self.rvr.rvr, index=index)

        move_pct = ((end_cur - px_start).abs()) / px_start
        mapping = {chg: srs[chg] for chg in move_pct.unique()}
        arr = move_pct.map(mapping).values

        if self.rvr.start_limit is not None:
            subset = arr[: self.rvr.start_limit[1]]
            min_limit = self.rvr.start_limit[0]
            subset[subset < min_limit] = min_limit
        if self.rvr.end_limit is not None:
            subset = arr[self.rvr.end_limit[1] :]
            max_limit = self.rvr.end_limit[0]
            subset[subset > max_limit] = max_limit
        return arr

    def _conf_end_rvr_pct(
        self,
        is_adv: bool,
        start: pd.Timestamp,
        conf_start: pd.Timestamp,
        end_alt: pd.Timestamp | None,
    ) -> tuple[pd.Timestamp | None, float | None, pd.Series, float, np.ndarray]:
        """Get info for when movement would be confirmed as ending by exceeding rvr.

        Parameters
        ----------
        end_alt
            Earliest date by which movement would end by any other means.

        Returns
        -------
        5-tuple of [pd.Timestamp | None, float | None, pd.Series, float, np.ndarray]
            [0] Bar on which movement would end by reversal. None if
                movement would not be considered to have given way to
                reversal as at `end_alt`.
            [1] Price at reversal limit (or nearest traded price to). None
                if movement would not be considered to have exceeded
                reversal limit as at end of available data.
            [2] Line representing reversal limit.
            [3] Maximum reversal within period before when would have, in
                any event, been considered to have ended by consolidation.
                If movement had not ended by end of available data then
                maximum reversal as at end of available data.
            [4] Array describing reversal limits. First value corresponds
                with reversal limit on bar when movement was confirmed.
        """
        row = self.data.loc[start]
        px_start = row.low if is_adv else row.high

        df = self.data.loc[conf_start:end_alt]
        end_cur = df.high.cummax() if is_adv else df.low.cummin()
        end_rvr = df.low.copy() if is_adv else df.high.copy()
        # for intervals where end_cur has been moved out, not possible to
        # know if the high/low came before the new low/high or afterwards,
        # given which assumes counter movement reached only interval close.
        bv = end_cur != end_cur.shift(1)
        end_rvr[bv] = df.close[bv]
        end_cur, end_rvr = end_cur[1:], end_rvr[1:]  # do not include conf_start

        chg = end_cur - px_start if is_adv else px_start - end_cur

        chg_rvr = (end_rvr - end_cur).abs()
        rvr = chg_rvr / chg
        rvr_max = rvr.max()

        rvr_arr = self._get_rvr_arr(is_adv, end_cur, px_start)
        line = px_start + (chg * ((1 - rvr_arr) * (1 if is_adv else -1)))
        bv = end_rvr < line if is_adv else end_rvr > line
        if not bv.any():
            return None, None, line, rvr_max, rvr_arr

        idx = bv.argmax()
        stop = idx + 1 if end_alt is None else min(idx + 1, len(bv))
        line_ = line.iloc[:stop]

        px_break = line.iloc[idx]
        row = df.iloc[idx + 1]
        px = min(px_break, row.open) if is_adv else max(px_break, row.open)
        return bv.index[idx], px, line_, rvr_max, rvr_arr

    def _conf_end_consol(
        self, is_adv: bool, conf_start: pd.Timestamp
    ) -> tuple[pd.Timestamp | None, float | None, pd.Series]:
        """Return bar when movement would be confirmed as ending by consolidation.

        Returns
        -------
        3-tuple of [pd.Timestamp | None, float | None, pd.Series]
            [0] Indice of `df` that movement would end by consolidation.
                None if movement ongoing as at end of available data.
            [1] Open price of bar when movement would end by consolidation.
                None if movement ongoing as at end of available data.
            [2] Line representing limit beyond which movement must be
                extended in order to be considered continuing as opposed to
                having given way to consolidation.
        """
        df = self.data.loc[conf_start:]
        col = df.high if is_adv else df.low
        win = col.rolling(self.prd, min_periods=0)
        fctrs = self.factors_break_pos_rev if is_adv else self.factors_break_neg_rev

        def rolling_win_func(arr: np.ndarray) -> float:
            win = arr * fctrs[-len(arr) :]
            return win.max() if is_adv else win.min()

        # line that, if not exceeded over prd, movement will be considered to have ended
        line_ = win.apply(rolling_win_func)
        # shift as evaluating limits for subsequent bar to exceed
        line = line_.shift(1, fill_value=line_.iloc[0])

        if len(df) <= self.prd:
            return None, None, line

        # end is first instance when self.prd consecutive bars fail to break line
        bv = (col > line if is_adv else col < line)[1:].rolling(self.prd).sum() == 0
        if not bv.any():
            return None, None, line
        end = bv.index[bv.argmax()]
        return end, df.loc[end].open, line[:end]

    def conf_end(
        self, start: pd.Timestamp, conf_start: pd.Timestamp, is_adv: bool
    ) -> tuple[
        pd.Timestamp | None,
        float | None,
        bool | None,
        bool | None,
        pd.Series,
        pd.Series,
        pd.Series | None,
        float,
        np.ndarray,
    ]:
        """Confirm movement end.

        Parameters
        ----------
        start
            Bar on which movement started.

        conf_start
            Bar when movement was confirmed.

        is_adv
            True: movement represents an advance.
            false: movement represents a decline.

        Returns
        -------
        9-tuple of
        [pd.Timestamp | None, float | None, bool | None, bool | None, pd.Series,
        pd.Series, pd.Series | None, float, np.ndarray]
            [0] pd.Timestamp | None
                Timestamp when movement's end confirmed.

                None if movement ongoing as at end of available data.

            [1] float | None
                If movement ended by consolidation, close price of bar when
                movement's end confirmed.

                If movement ended by recovery, limit price on breaking
                which movement was considered to have given way to
                recovery.

                None if movement ongoing as at end of available data.

            [2] bool | None
                True: movement ended by way of consolidation.
                False: movement ended by way of reversal.
                None: movement ongoing as at end of available data.

            [3] bool | None
                If movement ended by way of reversal, indicates nature of
                reversal.

                True: movement ended by way of price reversing more than
                    the reversal limit (i.e. more than rvr percent of the
                    trend movement).
                False: movement ended by way movement coming through in the
                    opposite direction.
                None: movement did not end by way of reversal.

            [4] pd.Series
                Line representing limit beyond which movement must be
                extended in order to be considered continuing as opposed
                to having given way to consolidation.

            [5] pd.Series
                Line representing reversal limit.

            [6] pd.Series | None
                If movement gave way to a movement in the opposite
                direction, the line indicating limit of consolidation range
                the breaking of which confirmed the opposing movement.

                None if movement did not give way to a movement in the
                opposite direction.

            [7] float
                Maximum percentage reversal before movement would have been
                considered to have given way to consolidation.

            [8] np.ndarray
                Array describing reversal limits. First value corresponds
                with reversal limit on bar when movement was confirmed.
        """
        end_rvr_opp, end_rvr_opp_px, line_rvr_opp = self._conf_end_rvr_opp(
            is_adv, conf_start
        )
        end_cnsl, end_cnsl_px, line_cnsl = self._conf_end_consol(is_adv, conf_start)

        tss = [ts for ts in [end_cnsl, end_rvr_opp] if ts is not None]
        end_alt = min(tss) if tss else None
        end_rvr, end_rvr_px, line_rvr_pct, rvr, rvr_arr = self._conf_end_rvr_pct(
            is_adv, start, conf_start, end_alt
        )

        lines = line_cnsl, line_rvr_pct
        rvrs = rvr, rvr_arr
        if end_alt is None and end_rvr is None:
            return None, None, None, None, *lines, None, *rvrs

        if end_rvr_opp is None:
            end_rvr_by_opp = False
        elif end_rvr is None:
            end_rvr_by_opp = True
        elif end_rvr_opp == end_rvr:
            # would be considered to have ended in both manners on same bar
            end_rvr_by_opp = (
                end_rvr_opp_px >= end_rvr_px if is_adv else end_rvr_opp_px <= end_rvr_px
            )
        else:
            end_rvr_by_opp = end_rvr_opp < end_rvr

        if end_rvr_by_opp:
            end_rvr = end_rvr_opp
            end_rvr_px = end_rvr_opp_px
        else:
            line_rvr_opp = None

        if end_cnsl is None:
            return (
                end_rvr,
                end_rvr_px,
                False,
                end_rvr_by_opp,
                *lines,
                line_rvr_opp,
                *rvrs,
            )
        if end_rvr is None or end_cnsl < end_rvr:
            return end_cnsl, end_cnsl_px, True, None, *lines, None, *rvrs
        return (end_rvr, end_rvr_px, False, end_rvr_by_opp, *lines, line_rvr_opp, *rvrs)

    def get_end_est_line(
        self, base: float, index: pd.DatetimeIndex, is_adv: bool
    ) -> pd.Series:
        """Get end establishment line (EEL).

        Parameters
        ----------
        base:
            Last y-value of EEL.

        index:
            x-values of EEL.

        is_adv:
            True if establishing end of an advance.
            False if establishing end of a decline.
        """
        factors = self.grad_factors_neg_rev if is_adv else self.grad_factors_pos_rev
        factors = factors[-len(index) :]
        return pd.Series(factors, index=index, dtype="float64") * base

    def get_movement_end(
        self, start: pd.Timestamp, end_conf: pd.Timestamp, is_adv: bool, by_consol: bool
    ) -> tuple[pd.Timestamp | pd.Interval, float, pd.Series | None]:
        """Get movement end bar and price.

        Parameters
        ----------
        start
            Bar on which movement started.

        end_conf
            Bar on which end of movement was confirmed.

        is_adv
            True: movement represents an advance.
            false: movement represents a decline.

        by_consol
            True: movement ended by way of consolidation.
            False: movement ended by way of reversal.

        Returns
        -------
        2-tuple of [pd.Timestamp | pd.Interval, float, pd.Series | None]
            [0] Bar representing end of movement.
            [1] Price as at movement end.
            [2] End Establishment Line if ended by consol, otherwise None.
        """
        df = self.data.loc[start:end_conf]

        idx = df.high.argmax() if is_adv else df.low.argmin()
        row = df.iloc[idx]
        end = row.name
        if not by_consol:
            return end, row.high if is_adv else row.low, None

        px = row.high if is_adv else row.low
        subset = df.iloc[max(0, idx - self.prd + 1) : idx]
        eel = self.get_end_est_line(px, subset.index, is_adv)
        bv = subset.high > eel if is_adv else subset.low < eel
        eel[row.name] = px
        if bv.any():
            idx_bv = subset[bv].high.argmax() if is_adv else subset[bv].low.argmin()
            end = subset[bv].index[idx_bv]
            return end, subset.loc[end].high if is_adv else subset.loc[end].low, eel
        return end, row.high if is_adv else row.low, eel

    def get_movement(
        self, df: pd.DataFrame, minstart_conf: pd.Timestamp, minstart: pd.Timestamp
    ) -> MovementAlt | None:
        """Define the first concluded movement.

        Defines the first movement which:
            is confirmed no earlier than `minstart_conf`.
            starts no earlier than `minstart`
            concludes before end of `df`.

        An advance/decline is considered to be confirmed on a bar when the
        y value is greater/lesser than all values of the prior
        "`self.prd` - 1" bars and has extened those prior highs/lows by a
        greater degree than would be considered as consolidation (i.e. as
        determined by having exceeded the Limit Line - see introduction to
        class doc).

        Returns None if no movement is identified.

        Parameters
        ----------
        df
            pd.DataFrame with rows representing bar interval and columns as
            OHLC. Method makes no changes to this dataframe.

        minstart_conf
            Returned movement will have a confirmation bar no earlier than
            `minstart_conf`.

        minstart
            Returned movement will have a start bar no earlier than
            `minstart`.
        """
        move = self.find_movement(df, minstart, minstart_conf)
        if move is None:
            return None
        is_adv, start, start_px, sel, start_conf, start_conf_px, start_conf_line = move

        (
            end_conf,
            end_conf_px,
            by_consol,
            rvr_by_opp,
            cns_line,
            rvr_pct_line,
            rvr_opp_line,
            rvr,
            rvr_arr,
        ) = self.conf_end(start, start_conf, is_adv)

        if end_conf is None:  # movement had not ended as at end of data
            end, eel = None, None
            subset = self.data.loc[start:]
            end_px = subset.high.max() if is_adv else subset.low.min()
            duration = len(df) - df.index.get_loc(start)
        else:
            end, end_px, eel = self.get_movement_end(start, end_conf, is_adv, by_consol)
            duration = df.index.get_loc(end) - df.index.get_loc(start) + 1

        movement = MovementAlt(
            is_adv,
            start,
            start_px,
            start_conf,
            start_conf_px,
            end,
            end_px,
            end_conf,
            end_conf_px,
            duration,
            self.params,
            sel,
            start_conf_line,
            eel,
            cns_line,
            rvr_pct_line,
            rvr_opp_line,
            by_consol,
            False if rvr_by_opp is None else not rvr_by_opp,
            rvr,
            rvr_arr,
        )
        return movement

    def get_movements(self) -> MovementsAlt:
        """Evaluate all movements."""
        moves = []
        df = self.data.copy()
        minstart_conf = df.index[self.prd]
        move = self.get_movement(df, minstart_conf, df.index[0])
        if move is None:
            return MovementsAlt(tuple(), self.data.copy(), self.interval)
        moves.append(move)
        end = move.end

        while True:
            end_idx = self.data.index.get_loc(end)
            df = self.data[end_idx - self.prd :].copy()
            if end_idx + 1 == len(self.data):
                return MovementsAlt(tuple(moves), self.data.copy(), self.interval)
            minstart_conf = self.data.index[end_idx + 1]
            move = self.get_movement(df, minstart_conf, end)
            if move is None:
                return MovementsAlt(tuple(moves), self.data.copy(), self.interval)
            # assert movement does not overlap with prior movement
            # lose assertion when test suite sufficient
            assert move.start >= moves[-1].end
            moves.append(move)
            if move.end is None:
                return MovementsAlt(tuple(moves), self.data.copy(), self.interval)
            end = move.end



"""Dictionay-related utility functions and classes"""

# pylint: disable=line-too-long


def set_kwargs_from_dflt(passed: dict, dflt: dict, deep: bool = False) -> dict:
    """Add default kwargs.

    Parameters
    ----------
        passed : dict
            Dictionary of kwargs to which default kwargs, not otherwise
            present, to be added.

        dflt : dict
            Dictionary of default kwargs.

        deep : bool, default: False
            If True will execute recurssively to add default kwargs to
            any value of passed that is itself a dictionary for
            which a corresponding dictionary in dflt exists.

    Examples
    --------
    >>> default = {
    ...     'one': 1,
    ...     'two': {
    ...         'two.one': 2.1,
    ...         'two.two': 2.2,
    ...     },
    ...     'three': 3
    ... }
    >>> passed = {
    ...     'one': 1.7,
    ...     'two': {
    ...         'two.one': '2.1.1',
    ...         'two.three': '2.3.1',
    ...     },
    ...     'four': 4
    ... }
    >>> set_kwargs_from_dflt(passed, default, deep=True)
    {'one': 1.7, 'two': {'two.one': '2.1.1', 'two.three': '2.3.1', 'two.two': 2.2}, 'four': 4, 'three': 3}
    """
    for key, value in dflt.items():
        if isinstance(value, dict) and deep and key in passed:
            set_kwargs_from_dflt(passed[key], value)
        passed.setdefault(key, value)
    return passed


def exec_kwargs(kwargs: dict) -> dict:
    """Update callable dictionary values to callable's return."""
    for k, v in kwargs.items():
        if callable(v):
            kwargs[k] = v()
    return kwargs


def set_kwargs_from_dflt_exec(passed: dict, dflt: dict) -> dict:
    """Update dictionary, including with callable values.

    Updates `passed` dictionary for missing items defined in `dflt`. Also
    sets value of any kwarg that represents a callable to the callable's
    return value (regardless of whether orinally defined in `passed` or
    `dflt`).
    """
    kwargs = set_kwargs_from_dflt(passed, dflt)
    return exec_kwargs(kwargs)


def update_deep(d: dict, u: dict, /) -> dict:
    """Updates dictionary to include any nested dictionaries.

    Code copied from Alex Telon / Alex Martelli response at:
        https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth.

    Parameters
    ----------
    d: dictionary to be updated.
    u: dictionary holding items with which to update d.

    Examples
    --------
    >>> default = {
    ...     'one': 1,
    ...     'two': {
    ...         'two.one': 2.1,
    ...         'two.two': 2.2,
    ...     },
    ...     'three': 3
    ... }
    >>> new = {
    ...     'one': 1.7,
    ...     'two': {
    ...         'two.one': '2.1.1',
    ...         'two.three': '2.3.1',
    ...     },
    ...     'four': 4
    ... }
    >>> update_deep(default, new)
    {'one': 1.7, 'two': {'two.one': '2.1.1', 'two.two': 2.2, 'two.three': '2.3.1'}, 'three': 3, 'four': 4}
    >>> default
    {'one': 1.7, 'two': {'two.one': '2.1.1', 'two.two': 2.2, 'two.three': '2.3.1'}, 'three': 3, 'four': 4}
    """
    for k, v in u.items():
        if isinstance(v, dict):
            d[k] = update_deep(d.get(k, {}), v)
        else:
            d[k] = v
    return d
